""" A subpackage that provides Python object persistence in HDF5 files.

It was written to make the storage of arbitrary nested Python structures
in the exchangable HDF5 format easy. Its main purpose is to easily add
persistence to existing numerical or data analysis codes.

While the files itself are plain HDF5 and can be read in any language
supporting HDF5, the format is not compatible to Matlab's own file format.
If you are searching for such a solution look at the hdf5storage package.

Usage
-----
The module exports a ``save()`` function that stores arbitrary structures
of Python and NumPy data types. For example

>>> x = {'data': [1, 2, 3], 'xrange': np.arange(5, dtype=np.uint8)}
>>> io.save(x, "test.hdf5")

This function should suffice for most needs as long as only standard types
are used. The ``load()`` function loads these files and restores the structure
and the types of the data:

>>> io.load("test.hdf5")
{'data': [1, 2, 3], 'xrange': array([0, 1, 2, 3, 4], dtype=uint8)}

Custom Objects
---------------
If you are using objects as simple containers without functionality you may
consider using the SimpleNamespace class from the ``types`` module of the
standard library. The advantage is that io knows how to handle it.::

    from types import SimpleNamespace
    a = SimpleNamespace(name="my object", data=np.arange(5))
    a.data2 = np.arange(10)
    copra.save(a)

If your objects are containers with methods but without a custom ``__init__()``
the simplest way is to inherit or mix-in the ``IO`` class::

    class Data(io.IO):
        x = 1

        def squared(self):
            return self.x * self.x

When using the ``IO`` class by default all instance attributes are stored
and loaded. More flexibility can be achieved by specifying ``_io``-attributes
of your custom class.

    _io_store : list of str or None, optional
        Specify the the instance attributes that are stored exclusively. Acts
        as a whitelist. If ``None`` all instance attributes are stored. Default
        is ``None``.
    _io_store_not : list of str or None, optional
        Specify which instance attributes are not stored. Acts as a blacklist.
        If ``None`` no blacklisting is done.

If you want to add attributes to storage you can call the
``_io_add_to_storage(key)`` method on your instance.
The IO class initalizes the instance without calling ``__init__()``. Instead
``__new__()`` is called on the class and afterwards the ``_post_init()``
method which subclasses can implement. A fully working example of a class
is the following (reduced from copra.FourierTransform)::

    class Grid(io.IO):
        _io_store = ['N', 'dx', 'x0']

        def __init__(self, N, dx, x0=0.0):
            # This is _not_ called upon loading from storage
            self.N = N
            self.dx = dx
            self.x0 = x0
            self._post_init()

        def _post_init(self):
            # this is called upon loading from storage
            # calculate the grids
            n = np.arange(self.N)
            self.x = self.x0 + n * self.dx

In this example the object can be exactly reproduced upon loading but only
a minimal amount of storage is required.

If you want to implement your own storage interface for a custom object
you should inherit from ``IO`` and implement your own ``to_dict()`` and
``from_dict()`` methods. Look at the implementation of the default in ``IO``
to understand their behavior.


File Format
-----------
The file format this module uses is a straightforward mapping of Python
types to the HDF5 data structure. Dictionaries and objects are mapped to
HDF5 groups, numpy arrays use h5py's type translation.
Iterables are converted to groups by introducing artificial keys of the
type ``idx_%d``. This is rather inefficient which explains why the
module should not be used to store large numerical arrays as a Python list.
To store the type information it uses an HDF5 attribute ``__class__``.
Furthermore, for scalars the attribute ``__dtype__`` and for strings the
attribute ``__encoding__`` are additionally used.

In conclusion, nested structures of Python types stored with this package are
not suitable for exchanging. Dictionaries of numerical data stored with this
package can be easily opened with any program that supports HDF5.
"""
from .handlers import (save_to_level, load_from_level, TypeHandler,
                       InstanceHandler)
from .io import save, load, IO, MetaIO



""" This module provides classes to calculate the refractive index
based on Sellmeier equations.

This is required to correctly model d-scan measurements.

Currently only very few materials are implemented. But more should be easy
to add. If the refractive index is described by formula 1 or 2 from
refractiveindex.info you can simply instantiate `SellmeierF1` or `SellmeierF2`.
If not, inherit from BaseMaterial and implement the `self._func` method.
"""
import numpy as np
from .frequencies import convert
from . import lib
from . import io


class BaseMaterial(io.IO):
    """ Abstract base class for dispersive materials.
    
    """
    
    def __init__(self, coefficients, freq_range, scaling=1.0e6,
                 check_bounds=True, name="", long_name=""):
        ''' Creates a dispersive material.
        
        Parameters
        ----------
        coefficients: ndarray
            The Sellmeier coefficients.
        freq_range : iterable
            The wavelength range in which the Sellmeier equation is valid
            (given in m).
        check_bounds : bool, optional
            Specifies if the frequency argument should be checked on every
            evaluation to match the allowed range.
        scaling : float, optional
            Specifies the scaling of the Sellmeier formula. E.g., most
            Sellmeier formulas are defined in terms of µm (micrometer), 
            whereas our function interface works in meter. In that case the
            scaling would be `1e6`. Default is `1.0e6`.
        '''
        if len(freq_range) != 2:
            raise ValueError("Frequency range must specified with two elements.")
        self._coefficients = np.array(coefficients)
        self._range = np.array(freq_range)
        self._scaling = scaling
        self.check = check_bounds
        self.name = name
        self.long_name = long_name
        
    def _check(self, x):
        if not self.check:
            return
        minx, maxx = np.min(x), np.max(x)
        if (minx < self._range[0]) or (maxx > self._range[1]):
            raise ValueError('Wavelength array [%e, %e] outside of valid range '
                             'of the Sellmeier equation [%e, %e].' % 
                             (minx, maxx, self._range[0], self._range[1]))          
    
    def _convert(self, x, unit):
        '''This is intended for conversion to be used in `self._func`.'''
        if unit != 'wl':
            x = convert(x, unit, 'wl')
        self._check(x)
        if self._scaling != 1.0:
            x = x * self._scaling
        return x    
    
    def n(self, x, unit='wl'):
        '''The refractive index at frequency `x` specified in units `unit`. '''
        return self._func(self._convert(x, unit))    
    
    def k(self, x, unit='wl'):
        '''The wavenumber in the material in rad / m.'''
        wl = convert(x, unit, "wl")
        return self.n(wl, unit="wl") * lib.twopi / wl


class SellmeierF1(BaseMaterial):
    ''' Defines a dispersive material via a specific Sellmeier equation.

        This subclass supports materials with a Sellmeier equation of the
        form::

            n^2(l) - 1 = c1 + c2 * l^2 / (l2 - c3^2) + ...

        This is formula 1 from refractiveindex.info [DispersionFormulas]_.
    '''
    def _func(self, x):
        c = self._coefficients
        x2 = x * x
        n2 = np.full_like(x, 1.0 + c[0])
        for i in range(1, len(c)-1, 2):
            n2 += c[i] * x2 / (x2 - c[i+1] * c[i+1])
        return np.sqrt(n2)


class SellmeierF2(BaseMaterial):
    ''' Defines a dispersive material via a specific Sellmeier equation.

        This subclass supports materials with a Sellmeier equation of the
        form::

            n^2(l) - 1 = c1 + c2 * l^2 / (l2 - c3) + ...

        This is formula 2 from refractiveindex.info [DispersionFormulas]_.
    '''
    def _func(self, x):
        c = self._coefficients
        x2 = x * x
        n2 = np.full_like(x, 1.0 + c[0])
        for i in range(1, c.size - 1, 2):
            n2 += c[i] * x2 / (x2 - c[i+1])
        return np.sqrt(n2)

FS = SellmeierF1(coefficients=[0.0000000, 0.6961663,
                               0.0684043, 0.4079426,
                               0.1162414, 0.8974794,
                               9.8961610],
                 freq_range=[0.21e-6, 6.7e-6],
                 name="FS",
                 long_name="Fused silica (fused quartz)")
"""Material instance describing fused silica (fused quartz).

The data was taken from refractiveindex.info
"""

BK7 = SellmeierF2(coefficients=[0.00000000000, 1.039612120,
                                0.00600069867, 0.231792344,
                                0.02001791440, 1.010469450,
                                103.560653],
                  freq_range=[0.3e-6, 2.5e-6],
                  name="BK7", long_name="N-BK7 (SCHOTT)")
"""Material instance describing N-BK7 (SCHOTT).

The data was taken from refractiveindex.info
"""


"""
Disclaimer
----------

THIS CODE IS FOR EDUCATIONAL PURPOSES ONLY! The code in this package was not
optimized for accuracy or performance. Rather it aims to provide a simple
implementation of the basic algorithms.

Author: Nils C. Geib, nils.geib@uni-jena.de
"""
__version__ = "0.1alpha"
from .autocorrelation import autocorrelation
from .fourier import FourierTransform
from .pulse import Pulse
from .random_pulse import random_pulse, random_gaussian
from .pulse_error import pulse_error
from .pnps import PNPS
from .mesh_data import MeshData
from .graphics import MeshDataPlot, PulsePlot
from .retrieval import Retriever
from . import lib
from . import material
from . import io
from .io import load, save



# -*- coding: utf-8 -*-
"""
Created on Sat May 25 20:50:09 2019

@author: Patrick
"""

import zipfile
import numpy as np
import femtoQ.tools as fq
import matplotlib.pyplot as plt
import scipy.signal as sig
import scipy.integrate as integrate
import scipy.interpolate as interp


C = 299792458

def NextPowerOfTwo(number):
    """ Returns next power of two following 'number' """
    return int(np.ceil(np.log2(number)))

def unzip(folder,filename, tempUnzipDir):
    """ Unzip content of Labview's output zipped data file  """
    
    with zipfile.ZipFile(folder+filename,"r") as zip_ref:
        zip_ref.extractall(folder+tempUnzipDir)
    
    return 0



def import_data(folder,filename, tempUnzipDir):
    """ Load data from binary files extracted from Labview
    zip file """
    
    binFilenames = [ folder+tempUnzipDir+'/'+filename[0:-4]+'_fixed.bin', folder+tempUnzipDir+'/'+filename[0:-4]+'_shear.bin', folder+tempUnzipDir+'/'+filename[0:-4]+'_scan.bin']
        
    N = np.fromfile(binFilenames[0], count = 2, dtype='>i')
    fixedMirrorData = np.fromfile(binFilenames[0], dtype='>d')[1:]
    fixedMirrorData = np.reshape(fixedMirrorData, newshape = ((N[0],N[1])))
    
    
    movingMirrorData = np.fromfile(binFilenames[1], dtype='>i')
    movingMirrorData = np.reshape(movingMirrorData[2:], newshape = (movingMirrorData[0],movingMirrorData[1]))
    movingMirrorData = movingMirrorData[:,1:]
    
    tdsiTrace = np.fromfile(binFilenames[2], dtype='>i')
    tdsiTrace = np.reshape(tdsiTrace[2:], newshape = (tdsiTrace[0],tdsiTrace[1]))
    tdsiTrace = tdsiTrace[:,1:]

    return fixedMirrorData, movingMirrorData, tdsiTrace


def cut_spectrum(minWavelength, maxWavelength, fixedMirrorData, movingMirrorData, wavelengths, tdsiTrace):
    """ Cut spectra to relevant wavelengths region """
    
    ii1 = np.argmin( np.abs(wavelengths - (minWavelength * 1e9 )))
    ii2 = np.argmin( np.abs(wavelengths - (maxWavelength * 1e9 )))


    fixedMirrorData = fixedMirrorData[ii1:ii2+1]
    movingMirrorData = movingMirrorData[:,ii1:ii2+1]
    wavelengths = wavelengths[ii1:ii2+1]
    tdsiTrace = tdsiTrace[:,ii1:ii2+1]
    
    return fixedMirrorData, movingMirrorData, wavelengths, tdsiTrace


def substractNoiseFloor(fixedMirrorData, movingMirrorData):
    """ Substract noise floor from spectra (except trace) """
    
    lowLimit = 0.01
    
    fixedMirrorData = fixedMirrorData - np.max((fixedMirrorData[0],fixedMirrorData[-1]))
    fixedMirrorData[fixedMirrorData<0] = 0
    maxFixed = np.max(fixedMirrorData)
    fixedMirrorData[ fixedMirrorData < maxFixed*lowLimit] = 0
    
    for ii in range(movingMirrorData.shape[0]):
        movingMirrorData[ii,:] = movingMirrorData[ii,:] - np.max((movingMirrorData[ii,0],movingMirrorData[ii,-1]))
        movingMirrorData[ii,:][movingMirrorData[ii,:]<0] = 0
        maxMoving = np.max(movingMirrorData[ii,:])
        movingMirrorData[ii,:][ movingMirrorData[ii,:] < maxMoving*lowLimit] = 0
    
    
    return fixedMirrorData, movingMirrorData


def find_shear(wavelengths, upconvPowerSpectrum, movingMirrorData, movingMirror_Z,tdsiTraceZ,debugGraphs):
    """ Calculate shear frequency as a function of stage position,
    and take its value at the middle position as constant approximation """
    
    # Normalize spectra to max of one
    upconvPowerSpectrum = upconvPowerSpectrum / np.max(upconvPowerSpectrum)
    
    maxValues = np.max(movingMirrorData, axis = 1)
    tmp = wavelengths.shape[0]
    maxValues = np.transpose( np.tile(maxValues, (tmp,1)) )
    movingMirrorData = movingMirrorData / maxValues
    


    # Convert wavelengths to frequencies
    frequencies = C / wavelengths
    
    frequencies = np.flip(frequencies) # Flipping from low to high frequencies
    upconvPowerSpectrum = np.flip(upconvPowerSpectrum)
    movingMirrorData = np.flip(movingMirrorData,axis = 1)
    
    # Interpolate to a linear spacing of frequencies
    # Choice of datapoint position strongly affect results, here I am copying Matlab
    # need to check if another strategy would work better
    Df = frequencies[-1] - frequencies[0]
    df = np.max( np.diff(frequencies) ) / 16
    N = int(round(Df / df))
    linFreqs = np.linspace(frequencies[-1]-(N-1)*df, frequencies[-1], N )
    upconvPowerSpectrum = np.interp(linFreqs, frequencies, upconvPowerSpectrum)
    
    
    newMovingMirrorData = np.zeros((movingMirrorData.shape[0], linFreqs.shape[0]))
    
    for ii in range( movingMirrorData.shape[0] ):
        newMovingMirrorData[ii,:] = np.interp(linFreqs, frequencies, movingMirrorData[ii,:])
    
    movingMirrorData = newMovingMirrorData
    frequencies = linFreqs
    
    
    
    #crossCorr = np.zeros( (movingMirrorData.shape[0], movingMirrorData.shape[1]))
    crossCorr = np.zeros( (movingMirrorData.shape[0], movingMirrorData.shape[1]*2-1))

    lags = np.zeros_like(crossCorr)
    shearMap = np.zeros( movingMirrorData.shape[0] )
    
    for ii in range( movingMirrorData.shape[0] ):
        #lags[ii,:], crossCorr[ii,:] =fq.ezcorr(frequencies, movingMirrorData[ii,:], upconvPowerSpectrum) 
        #shearMap[ii] = lags[ii,:][ crossCorr[ii,:] == np.max(crossCorr[ii,:]) ]
        crossCorr[ii,:] =  np.correlate(movingMirrorData[ii,:], upconvPowerSpectrum,'full')
        maxId = np.argmax(crossCorr[ii,:])
        peakFreq = -(N - (maxId+1))*df
        lags = -(N - np.linspace(1,2*N-1,2*N-1) ) *df
        
        
        x,y = fq.ezdiff(lags, crossCorr[ii,:])
        
        f = interp.interp1d(x, y, kind = 'cubic')
        
        err = 1
        threshold = 1e-5
        maxIter = 1000
        nIter = 0
        x0 = peakFreq - 5*df
        x1 = peakFreq + 5*df
            
        while err > threshold:
            nIter += 1
            if nIter > maxIter:
                break
                
            if x0<np.min(x):
                x0 = np.min(x)
            if x0>np.max(x):
                x0 = np.max(x)
                
            if x1<np.min(x):
                x1 = np.min(x)
            if x1>np.max(x):
                x1 = np.max(x)
                    
            f0 = f(x0)
            f1 = f(x1)
            dfdx = (f1-f0) / (x1 - x0)
            b = f0 - dfdx*x0
            
            x0 = x1
            x1 = -b/dfdx
            err = abs((x1-x0)/x0)
            
             
        
        shearMap[ii] = x1# -(N - (maxId+1))*df
    
    p = np.polyfit(movingMirror_Z, shearMap, 1)
    
    shear = np.polyval(p, (tdsiTraceZ[0]+tdsiTraceZ[-1])/2)
        
    if debugGraphs:
        plt.figure()
        plt.plot(movingMirror_Z,shearMap/1e12, LineWidth = 2, label = 'Data')
        plt.plot(movingMirror_Z, np.polyval(p, movingMirror_Z) /1e12, 'r', LineWidth = 2, label = 'Linear fit')
        plt.legend()
        plt.xlabel(r'Stage displacement [$\mu$m]')
        plt.ylabel('Shear [THz]')
        
        plt.figure()
        plt.pcolor(lags/1e12, movingMirror_Z, crossCorr)
        plt.plot(shearMap/1e12, movingMirror_Z, '--r')
        plt.xlim(-20,20)
        plt.xlabel('Shear frequency [THz]')
        plt.ylabel(r'Stage displacement [$\mu$m]')
    
    return shear


def find_upconversion(wavelengths, upconvPowerSpectrum, fundWavelengths, fundPowerSpectrum):
    """ Calculate shear frequency as a function of stage position,
    and take its value at the middle position as constant approximation """
    
    # Normalize spectra to max of one
    upconvPowerSpectrum = upconvPowerSpectrum / np.max(upconvPowerSpectrum)
    
    fundPowerSpectrum = fundPowerSpectrum/fundPowerSpectrum.max()


    # Convert wavelengths to frequencies
    frequencies = C / wavelengths
    fundFrequencies = C/ fundWavelengths
    
    frequencies = np.flip(frequencies) # Flipping from low to high frequencies
    fundFrequencies = np.flip(fundFrequencies)
    upconvPowerSpectrum = np.flip(upconvPowerSpectrum)
    fundPowerSpectrum = np.flip(fundPowerSpectrum)
    
    # Interpolate to a linear spacing of frequencies
    # Choice of datapoint position strongly affect results, here I am copying Matlab
    # need to check if another strategy would work better
    Df = np.min( [ np.min(np.diff(frequencies)),  np.min(np.diff(fundFrequencies))  ]  )
    maxFreq = frequencies.max()
    minFreq = fundFrequencies.min()
    
    N = int(round((maxFreq -minFreq) / Df))
    linFreqs = np.linspace(maxFreq, minFreq, N )
    
    upconvPowerSpectrum = np.interp(linFreqs, frequencies, upconvPowerSpectrum)
    fundPowerSpectrum = np.interp(linFreqs, fundFrequencies, fundPowerSpectrum)
    
    frequencies = linFreqs
    
    
    
    #crossCorr = np.zeros( (movingMirrorData.shape[0], movingMirrorData.shape[1]))
    crossCorr = np.zeros( linFreqs.shape[0]*2-1 )

    lags = np.zeros_like(crossCorr)
    
    crossCorr =  np.correlate(fundPowerSpectrum, upconvPowerSpectrum,'full')
    maxId = np.argmax(crossCorr)
    peakFreq = -(N - (maxId+1))*Df
    lags = -(N - np.linspace(1,2*N-1,2*N-1) ) *Df
    
    
    x,y = fq.ezdiff(lags, crossCorr)
    
    f = interp.interp1d(x, y, kind = 'cubic')
    
    err = 1
    threshold = 1e-5
    maxIter = 1000
    nIter = 0
    x0 = peakFreq - 5*Df
    x1 = peakFreq + 5*Df
        
    while err > threshold:
        nIter += 1
        if nIter > maxIter:
            break
            
        if x0<np.min(x):
            x0 = np.min(x)
        if x0>np.max(x):
            x0 = np.max(x)
            
        if x1<np.min(x):
            x1 = np.min(x)
        if x1>np.max(x):
            x1 = np.max(x)
                
        f0 = f(x0)
        f1 = f(x1)
        dfdx = (f1-f0) / (x1 - x0)
        b = f0 - dfdx*x0
        
        x0 = x1
        x1 = -b/dfdx
        err = abs((x1-x0)/x0)
        
         
    
    upconvWavelength = C/x1# -(N - (maxId+1))*df
    
    
    return upconvWavelength

def make1Dfft(wavelengths,stagePosition,trace,zeropadding = True, windowing = True, debugGraphs= False):

    #stagePosition = np.abs(stagePosition)
    
    dz = np.diff(stagePosition)[0]
    
    
    plt.figure()
    plt.pcolor(wavelengths*1e9,stagePosition,np.abs(trace)/np.max(np.abs(trace)))
    plt.xlabel(r'Wavelengths [nm]')
    plt.ylabel(r'Stage displacement [$\mu$m]')
    c = plt.colorbar()
    c.set_label('Relative intensity')
    
    
    if windowing:
        hanningWindow = np.hanning(stagePosition.shape[0])
        hanningWindow = np.transpose( np.tile(hanningWindow, (wavelengths.shape[0],1)))
        trace = trace * hanningWindow
        
    if zeropadding:
        deficit = 2**NextPowerOfTwo(stagePosition.shape[0]) - stagePosition.shape[0]
        deficitUp = int(np.floor(deficit/2))
        deficitDown = int(np.ceil(deficit/2))
        
        zeroPadUp = np.zeros((deficitUp,wavelengths.shape[0]))
        zeroPadDown = np.zeros((deficitDown,wavelengths.shape[0]))
        trace = np.concatenate((zeroPadUp, trace, zeroPadDown))
    
    
    
    # Perform 1D fft of 2DSI trace along stage position dimension
    fTrace = np.fft.fftshift(np.fft.fft(trace, axis = 0), axes = 0)
    
    # Calculate spatial frequencies axis
    spatialFreqs = np.fft.fftshift( np.fft.fftfreq(fTrace.shape[0], dz) )
    
    # Keep only positive frequencies
    iiMid = spatialFreqs>=0
    fTrace = fTrace[iiMid,:]
    spatialFreqs = spatialFreqs[iiMid]
    
    # Take absolute value of interferogram and find peak corresponding to spectral
    # phase oscillations
    absfTrace = np.abs( fTrace )
    
    
    tmp = np.mean(absfTrace, axis = 1)
    II = sig.find_peaks(tmp)[0]
    II = II[tmp[II] == np.max(tmp[II])]
    
    phase = np.squeeze( np.unwrap( np.angle( fTrace[II,:] ) ) )
    amplitude = np.squeeze( absfTrace[II,:] )
    spatialFreq = spatialFreqs[II]
    
    if debugGraphs:
        plt.figure()
        plt.pcolor(wavelengths*1e9, spatialFreqs, absfTrace/np.max(absfTrace))
        plt.xlabel('Wavelengths [nm]')
        plt.ylabel(r'Spatial frequency [$\mu$m$^{-1}$]')
        c = plt.colorbar()
        c.set_label('Relative intensity')
    
    return spatialFreq, amplitude, phase
               
               
def cut_fft(amplitude, phase, wavelengths,debugGraphs):
    
    cut = 0.00
    extend = 0
    
    Icut = amplitude >= cut*np.max(amplitude)
    Imin = np.argwhere(Icut)[0]
    Imax = np.argwhere(Icut)[-1]
    Idelta = (Imax - Imin)*extend
    
    Imin = np.squeeze( np.ceil(Imin - Idelta).astype(int) )
    Imax = np.squeeze( np.floor(Imax + Idelta).astype(int) )
    
    Icut = np.zeros_like(Icut, dtype = int)
    Icut[Imin : Imax+1] = 1
    IcutBool = Icut.astype(bool)
    #IcutIndex = np.squeeze( np.argwhere(Icut) )
    
    if debugGraphs:
        fig = plt.figure()
        ax = fig.gca()
        ax.plot(wavelengths*1e9, phase, label = 'Full data')
        
    amplitude = amplitude[IcutBool]
    phase = phase[IcutBool]
    FFTwavelengths = wavelengths[IcutBool]
    
               
    amplitude = amplitude / np.max(amplitude)
               
    
    phase = phase - np.fix( np.min( phase/(2*np.pi) ) )*2*np.pi
        
    if debugGraphs:       
        ax.plot(FFTwavelengths*1e9, phase,'--', label = 'Trunc. data')
        ax.set_xlabel('Wavelength [nm]')
        ax.set_ylabel('2DSI fringes phase [rad]')
        ax.legend()
    return amplitude, phase, FFTwavelengths
               
               
               

def calc_spectral_phase(shearFrequency, upconvWavelength,FFTamplitude, FFTphase, FFTwavelengths,polyfit,debugGraphs):
    
    
    shearW = 2*np.pi * shearFrequency
    
    phase  = -FFTphase # Minus to "correct GVD sign"...
    
    concW, concPhase,concGD, concGDD, concTOD = phase_concatenation(FFTwavelengths,FFTamplitude, shearW, phase, upconvWavelength,polyfit=polyfit ,debugGraphs =debugGraphs)
    midpointW, midpointPhase, midpointGD, midpointGDD, midpointTOD = phase_midpoint(FFTwavelengths,FFTamplitude, shearW, phase, upconvWavelength,polyfit=polyfit ,debugGraphs =debugGraphs)
    
    return concW, concPhase,concGD, concGDD, concTOD, midpointW, midpointPhase,midpointGD, midpointGDD, midpointTOD
    
    
    
def phase_concatenation(wavelengths,amplitude, shearW, phase, upconvWavelength, polyfit = 0,debugGraphs= False):
    
    wCut = 2*np.pi * C *( 1/wavelengths - 1/upconvWavelength)
    
    dwCut = np.mean( np.diff( wCut ) )
    
    shearMultiplicity = np.round( np.abs( shearW / dwCut ) ).astype(int)
    
    newDW = np.abs( shearW / shearMultiplicity)
    
    N = np.floor( (np.max(wCut) - np.min(wCut)) / newDW ).astype(int)
    
    wMin = np.max(wCut) - N*newDW
    
    wCutLin = np.linspace( wMin , np.max(wCut), N+1 )
    
    phaseCutLin = np.interp(wCutLin, np.flip(wCut), np.flip(phase))
    amplitudeCutLin = np.interp(wCutLin, np.flip(wCut), np.flip(amplitude))
    
    
    concPhase = np.zeros((shearMultiplicity, np.floor(phaseCutLin.shape[0]/shearMultiplicity).astype(int)+1 ))
    concW = np.zeros((shearMultiplicity, np.floor(phaseCutLin.shape[0]/shearMultiplicity).astype(int) ))
    
    
    for ii in range(np.floor(phaseCutLin.shape[0]/shearMultiplicity).astype(int)):
        
        concPhase[:, ii+1] = (concPhase[:, ii].transpose() - phaseCutLin[(ii)*shearMultiplicity : (ii+1)*shearMultiplicity]).transpose()
        concW[:, ii]= wCutLin[(ii)*shearMultiplicity : (ii+1)*shearMultiplicity].transpose()
    
    concPhase = concPhase[:,1:]
    
    if shearW < 0:
        concPhase = -concPhase
    
    concGD = np.zeros_like(concPhase[:,1:-1])
    concGDD = np.zeros_like(concPhase[:,1:-1])
    concTOD = np.zeros_like(concPhase[:,2:-2])
    
    for ii in range(shearMultiplicity):
        
        
        weights = np.interp(concW[ii,:], wCutLin, amplitudeCutLin)
        linFit = np.polyfit(concW[ii,:], concPhase[ii,:], 1, w = weights)
        
        concPhase[ii,:] = concPhase[ii,:] - np.polyval(linFit, concW[ii,:])
    
        if polyfit >= 2:
            phaseFit =  np.polyfit(concW[ii,:], concPhase[ii,:], polyfit, w = weights)
            concPhase[ii,:] = np.polyval(phaseFit, concW[ii,:])
        
        concGD[ii,:] = fq.ezdiff(concW[ii,:], concPhase[ii,:], n=1, order = 2)[1]
        concGDD[ii,:] = fq.ezdiff(concW[ii,:], concPhase[ii,:], n=2, order = 2)[1]
        concTOD[ii,:] = fq.ezdiff(concW[ii,:], concPhase[ii,:], n=3, order = 2)[1]
    
    if debugGraphs:
        plt.figure()
        plt.plot((2*np.pi*C/concW.transpose()*1e9), concPhase.transpose())
        plt.ylabel('Spectral phase (concat. method) [rad]')
        plt.xlabel('Downconv. wavelength [nm]')
        
    return concW, concPhase, concGD, concGDD, concTOD
        

def phase_midpoint(wavelengths,amplitude, shearW, phase, upconvWavelength, polyfit = 0,debugGraphs= False):
    
    wCut = 2*np.pi * C *( 1/wavelengths - 1/upconvWavelength)
    
    dwCut = np.mean( np.diff( wCut ) )
    
    shearMultiplicity = np.round( np.abs( shearW / dwCut ) ).astype(int)
    
    newDW = np.abs( shearW / shearMultiplicity)
    
    N = np.floor( (np.max(wCut) - np.min(wCut)) / newDW ).astype(int)
    
    wMin = np.max(wCut) - N*newDW
    
    wCutLin = np.linspace( wMin , np.max(wCut), N+1 )
    
    phaseCutLin = np.interp(wCutLin, np.flip(wCut), np.flip(phase))
    amplitudeCutLin = np.interp(wCutLin, np.flip(wCut), np.flip(amplitude))
    
    
    
    deltaOfW = integrate.cumtrapz(phaseCutLin, wCutLin)
    midpointW = wCutLin[1:]
    
    midpointPhase = -np.interp(midpointW-shearW/2, midpointW, deltaOfW) / shearW
   
    
    
    
    
    midpointGD = np.zeros_like(midpointPhase[1:-1])
    midpointGDD = np.zeros_like(midpointPhase[1:-1])
    midpointTOD = np.zeros_like(midpointPhase[2:-2])
    
    weights = np.interp(midpointW, wCutLin, amplitudeCutLin)
    linFit = np.polyfit(midpointW, midpointPhase, 1, w = weights)
        
    midpointPhase = midpointPhase - np.polyval(linFit, midpointW)
    
    if polyfit >= 2:
        phaseFit =  np.polyfit(midpointW, midpointPhase, polyfit, w = weights)
        midpointPhase = np.polyval(phaseFit, midpointW)
    
    
    midpointGD = fq.ezdiff(midpointW, midpointPhase, n=1, order = 2)[1]
    midpointGDD = fq.ezdiff(midpointW, midpointPhase, n=2, order = 2)[1]
    midpointTOD = fq.ezdiff(midpointW, midpointPhase, n=3, order = 2)[1]
    
    if debugGraphs:
        plt.figure()
        plt.plot((2*np.pi*C/midpointW.transpose()*1e9), midpointPhase.transpose())
        plt.ylabel('Spectral phase (midpoint method) [rad]')
        plt.xlabel('Downconv. wavelength [nm]')
        
    return midpointW, midpointPhase,midpointGD, midpointGDD, midpointTOD


def calc_temporal_envelope(wavelengths, upconvPowerSpectrum, upconvWavelength, angFreqPhase, phase, plotSpectrum,minTimeResolution):
    
    pulse = []
    
    if len(phase.shape) > 1:
    
        for ii in range(phase.shape[0]):
        
            frequency = np.flip( C * (1/wavelengths - 1/upconvWavelength) )
            
            spectrumEnvelope = np.flip( np.sqrt(upconvPowerSpectrum/np.max(upconvPowerSpectrum)) )
            
            nuMax = np.max(frequency)
            nuMin = np.min(frequency)
            dNu = np.min( np.diff( frequency ) )
            
            if minTimeResolution < 1/(2*nuMax):
                newNuMax = 1 / (2* minTimeResolution)
                nu = np.linspace(-newNuMax, newNuMax, int(np.floor(2*newNuMax/dNu))+1)
            else:
                nu = np.linspace(-nuMax, nuMax, int(np.floor(2*nuMax/dNu))+1)
            
            interpEnvelope = np.zeros_like(nu)
            interpPhase = np.zeros_like(nu)
            
            II = ( (nu>=nuMin) & (nu<=nuMax) )
            
            interpEnvelope[II] = np.interp(nu[II], frequency, spectrumEnvelope)
            interpPhase[II] = np.interp(nu[II], angFreqPhase[ii,:]/(2*np.pi), phase[ii,:])
            interpPhase[nu<nuMin] = interpPhase[II][0]
            interpPhase[nu>nuMax] = interpPhase[II][-1]
            
            spectrum = interpEnvelope * np.exp(1j * interpPhase)
            
            t, eField = fq.ezifft(nu, spectrum)
            eField = np.fft.ifftshift(eField)
            
            pulse.append( np.abs(eField) )
# =============================================================================
#             if (plotSpectrum and ii==0):
#                 IIplotphase = interpEnvelope[nu>0] > np.max(interpEnvelope)/10
#                 
#                 
#                 fig = plt.figure()
#                 axLeft = fig.gca()
#                 axRight = axLeft.twinx()
#                 axLeft.plot(C/nu[nu>0]*1e9, interpEnvelope[nu>0]**2)
#                 axRight.plot(C/nu[nu>0][IIplotphase]*1e9,interpPhase[nu>0][IIplotphase], '--r')
#                 axLeft.set_xlim(np.min(C/nu[nu>0]*1e9),1200)
#                 axLeft.set_xlabel('Wavelength [nm]')
#                 axLeft.set_ylabel('Normalized intensity', color = 'blue')
#                 axRight.set_ylabel('Spectral phase [rad]', color = 'red')
# =============================================================================
    
    else:
        frequency = np.flip( C * (1/wavelengths - 1/upconvWavelength) )
            
        spectrumEnvelope = np.flip( np.sqrt(upconvPowerSpectrum/np.max(upconvPowerSpectrum)) )
        
        nuMax = np.max(frequency)
        nuMin = np.min(frequency)
        dNu = np.min( np.diff( frequency ) )
        
        if minTimeResolution < 1/(2*nuMax):
                newNuMax = 1 / (2* minTimeResolution)
                nu = np.linspace(-newNuMax, newNuMax, int(np.floor(2*newNuMax/dNu))+1)
        else:
                nu = np.linspace(-nuMax, nuMax, int(np.floor(2*nuMax/dNu))+1)
            
        interpEnvelope = np.zeros_like(nu)
        interpPhase = np.zeros_like(nu)
        
        II = ( (nu>=nuMin) & (nu<=nuMax) )
        
        interpEnvelope[II] = np.interp(nu[II], frequency, spectrumEnvelope)
        interpPhase[II] = np.interp(nu[II], angFreqPhase/(2*np.pi), phase)
        interpPhase[nu<nuMin] = interpPhase[II][0]
        interpPhase[nu>nuMax] = interpPhase[II][-1]
        
        spectrum = interpEnvelope * np.exp(1j * interpPhase)
        
        t, eField = fq.ezifft(nu, spectrum)
        eField = np.fft.ifftshift(eField)
        
        pulse.append( np.abs(eField) )
# =============================================================================
#         if (plotSpectrum):
#             IIplotphase = interpEnvelope[nu>0] > np.max(interpEnvelope)/10
#             
#             
#             fig = plt.figure()
#             axLeft = fig.gca()
#             axRight = axLeft.twinx()
#             axLeft.plot(C/nu[nu>0]*1e9, interpEnvelope[nu>0]**2)
#             axRight.plot(C/nu[nu>0][IIplotphase]*1e9,interpPhase[nu>0][IIplotphase], '--r')
#             axLeft.set_xlim(np.min(C/nu[nu>0]*1e9),1200)
#             axLeft.set_xlabel('Wavelength [nm]')
#             axLeft.set_ylabel('Normalized intensity', color = 'blue')
#             axRight.set_ylabel('Spectral phase [rad]', color = 'red')
# =============================================================================
        
    
    pulse = np.array(pulse)
    pulse = np.mean(pulse, axis = 0)
    pulse = pulse**2 / np.max(pulse**2)
    
    
    
    return t, pulse, eField
    



def make_shear_map(wavelengths, upconvPowerSpectrum, movingMirrorData, movingMirror_Z,debugGraphs):
    """ Calculate shear frequency as a function of stage position,
    and take its value at the middle position as constant approximation """
    
    # Normalize spectra to max of one
    upconvPowerSpectrum = upconvPowerSpectrum / np.max(upconvPowerSpectrum)
    
    maxValues = np.max(movingMirrorData, axis = 1)
    tmp = wavelengths.shape[0]
    maxValues = np.transpose( np.tile(maxValues, (tmp,1)) )
    movingMirrorData = movingMirrorData / maxValues
    


    # Convert wavelengths to frequencies
    frequencies = C / wavelengths
    
    frequencies = np.flip(frequencies) # Flipping from low to high frequencies
    upconvPowerSpectrum = np.flip(upconvPowerSpectrum)
    movingMirrorData = np.flip(movingMirrorData,axis = 1)
    
    # Interpolate to a linear spacing of frequencies
    # Choice of datapoint position strongly affect results, here I am copying Matlab
    # need to check if another strategy would work better
    Df = frequencies[-1] - frequencies[0]
    df = np.max( np.diff(frequencies) ) / 2 #Added this division by two as a test
    N = round(Df / df)
    linFreqs = np.linspace(frequencies[-1]-(N-1)*df, frequencies[-1], N )
    upconvPowerSpectrum = np.interp(linFreqs, frequencies, upconvPowerSpectrum)
    
    
    newMovingMirrorData = np.zeros((movingMirrorData.shape[0], linFreqs.shape[0]))
    
    for ii in range( movingMirrorData.shape[0] ):
        newMovingMirrorData[ii,:] = np.interp(linFreqs, frequencies, movingMirrorData[ii,:])
    
    movingMirrorData = newMovingMirrorData
    frequencies = linFreqs
    
    
    
    #crossCorr = np.zeros( (movingMirrorData.shape[0], movingMirrorData.shape[1]))
    crossCorr = np.zeros( (movingMirrorData.shape[0], movingMirrorData.shape[1]*2-1))

    lags = np.zeros_like(crossCorr)
    shearMap = np.zeros( movingMirrorData.shape[0] )
    
    for ii in range( movingMirrorData.shape[0] ):
        #lags[ii,:], crossCorr[ii,:] =fq.ezcorr(frequencies, movingMirrorData[ii,:], upconvPowerSpectrum) 
        #shearMap[ii] = lags[ii,:][ crossCorr[ii,:] == np.max(crossCorr[ii,:]) ]
        crossCorr[ii,:] =  np.correlate(movingMirrorData[ii,:], upconvPowerSpectrum,'full')
        maxId = np.argmax(crossCorr[ii,:])
        
        lags = -(N - np.linspace(1,2*N-1,2*N-1) ) *df
        peakFreq = np.average(lags, weights = crossCorr[ii,:])
        
        
        shearMap[ii] = peakFreq#-(N - (maxId+1))*df
    
    p = np.polyfit(movingMirror_Z, shearMap, 1)
    
    shearMap = np.polyval(p, movingMirror_Z)
        
    if debugGraphs:
        plt.figure()
        plt.plot(movingMirror_Z/1000,shearMap/1e12, LineWidth = 2, label = 'Data')
        plt.plot(movingMirror_Z/1000, np.polyval(p, movingMirror_Z) /1e12, 'r', LineWidth = 2, label = 'Linear fit')
        plt.legend()
        plt.xlabel(r'Stage displacement [$\mu$m]')
        plt.ylabel('Shear [THz]')
    
    return shearMap


def simFrogTrace(t,E,minWavelength,maxWavelength):
    
    I = np.abs(E)**2
    
    t0 = -200e-15#t[t<0][I[t<0]<np.max(I)/100][-1]*2
    t1 = 200e-15#t[t>0][I[t>0]<np.max(I)/100][0]*2
    
    if abs(t1) > abs(t0):
        t0 = -t1
    else:
        t1 = -t0
    
    lags = np.linspace(t0, t1, 400)
    
    wavelengths = np.linspace(minWavelength, maxWavelength, 1000)
    frogTrace = np.zeros((wavelengths.shape[0],lags.shape[0]))
    
    for ii,lag in enumerate(lags):
        
        t2 = t + lag
        E2 = np.interp(t,t2,E)
        
        Esfg = E * E2
        
        
        nu, S = fq.ezfft(t, Esfg)
        
        wav = C/np.flip(nu)
        S = np.flip(S)
        spectrum = np.interp(wavelengths, wav, np.abs(S)**2)
        
        frogTrace[:,ii] = spectrum
    
    frogTrace = frogTrace / np.max(frogTrace)
    
    plt.figure()
    plt.pcolor(lags*1e15, wavelengths*1e9, frogTrace, cmap = 'viridis')
    plt.gca().invert_yaxis()
    plt.xlabel('Delay [fs]')
    plt.ylabel('Wavelength [nm]')
    c = plt.colorbar()
    c.set_label('Relative intensity')
    
    
        
        
    
    


""" This module implements retrieval algorithms based on general
nonlinear optimization algorithms such as Levenberg-Marquadt,
differential evolution, or Nelder-Mead.
"""
import numpy as np
from scipy import optimize
from .retriever import BaseRetriever


class NLORetriever(BaseRetriever):

    def _scalar_objective(self, x):
        # rename
        rs = self._retrieval_state
        log = self.log
        # calculate trace error
        En = x.view(np.complex128)
        r = self._objective_function(En)
        R = self._Rr(r)
        # printing and logging
        if rs.nfev % 100 == 0 and self.verbose:
            print(rs.nfev, R)
        if self.logging:
            log.trace_error.append(R)
        rs.nfev += 1
        # normalize the error to avoid ill-scaling
        return r / self.Tmn_meas.max()**2

    def _vector_objective(self, x):
        # rename
        rs = self._retrieval_state
        log = self.log
        # calculate the error vector
        En = x.view(np.complex128)
        Tmn = self.pnps.calculate(En, self.parameter)
        diff = self._error_vector(Tmn, store=False)
        R = self._Rr(np.sum(diff * diff))
        # printing and logging
        if rs.nfev % 100 == 0 and self.verbose:
            print(rs.nfev, R)
        if self.logging:
            log.trace_error.append(R)
        rs.nfev += 1
        # normalize the error vector to avoid ill-scaling
        return diff / np.sqrt(self.M * self.N) / self.Tmn_meas.max()

    def _retrieve_begin(self, measurement, initial_guess, weights):
        super()._retrieve_begin(measurement, initial_guess, weights)
        rs = self._retrieval_state
        rs.nfev = 0

    def _retrieve_end(self):
        super()._retrieve_end()
        self._result.nfev = self._retrieval_state.nfev

    def result(self, pulse_original=None, full=True):
        res = super().result(pulse_original=pulse_original, full=full)
        res.nfev = self._retrieval_state.nfev
        return res


class LMRetriever(NLORetriever):
    """ Implements pulse retrieval based on the Levenberg-Marquadt algorithm.

    This is an efficient nonlinear least-squares solver, however, it will still
    be *very* slow for large pulses (N > 256). The reason is that the
    (MN x N) Jacobian is evaluated using numerical differentiation.

    The recommendation is to use this method either on small problems or to
    refine or verify solutions provided by a different algorithm.
    """
    method = "lm"

    def __init__(self, pnps, ftol=1e-08, xtol=1e-08, gtol=1e-08, lm_verbose=0,
                 **kwargs):
        """ For a full documentation of the arguments see :class:`Retriever`.

        For the documentation of `ftol`, `xtol`, `gtol` see the documentation
        of :func:`scipy.optimize.least_squares`. They are passed directly
        to the optimizer. If you want to run the optimizer for a fixed
        number of iterations, set all values to 1e-14 to effectively
        disable the stopping criteria.
        """
        super().__init__(pnps, ftol=ftol, xtol=xtol, gtol=gtol,
                         lm_verbose=lm_verbose, **kwargs)

    def _retrieve(self):
        # local rename
        o = self.options
        res = self._result
        # store current guess in attribute
        spectrum = self.initial_guess.copy()
        # This algorithm is not robust against the scaling of the input vector!
        spectrum /= np.abs(spectrum).max()
        x0 = spectrum.view(np.float64).copy()
        # calculate the maximum number of function evaluations
        max_nfev = None
        if o.maxfev is not None:
            max_nfev = o.maxfev // x0.shape[0]
        optres = optimize.least_squares(
                        self._vector_objective,
                        x0,
                        method='trf',
                        jac='2-point',
                        max_nfev=max_nfev,
                        tr_solver='exact',
                        ftol=o.ftol,
                        gtol=o.gtol,
                        xtol=o.xtol,
                        verbose=o.lm_verbose
                    )
        res.approximate_error = False
        res.spectrum = optres.x.view(dtype=np.complex128)
        res.trace_error = self.trace_error(res.spectrum)
        res.approximate_error = False
        return res.spectrum


class DERetriever(NLORetriever):
    """ This retriever uses the gradient-free differential evolution algorithm.

    It tries to match the parameters described in [Escoto2018]_ as far as
    they are mentioned. No further effort was made to optimize them. If you
    are interested in using DE as a pulse retrieval algorithm you are
    advised to study the documentation at
    :func:`scipy.optimize.differential_evolution`.

    The initial population in our implementation is based on the provided guess
    with added complex, Gaussian noise of 5% of the maximum amplitude.
    In our tests we saw no convergence when starting from completely
    random initial guesses.
    """
    method = "de"

    def _retrieve(self):
        # local rename
        o = self.options
        res = self._result
        # calculate the maximum number of function evaluations
        max_nfev = None
        if o.maxfev is not None:
            max_nfev = int(round(o.maxfev / 10 - 1))
        # generate initial population
        init = [self.initial_guess.view(np.float64).copy()]
        amp = np.abs(self.initial_guess).max()
        for i in range(9):
            sol = (self.initial_guess +
                   0.05 * amp * np.random.normal(size=self.N) +
                   0.05j * amp * np.random.normal(size=self.N))
            init.append(sol.view(np.float64))
        optres = optimize.differential_evolution(
                    self._scalar_objective,
                    [(-1.0, 1.0) for i in range(2 * self.N)],
                    strategy='rand1bin',
                    maxiter=max_nfev,
                    recombination=0.5,
                    popsize=10,  # is overwritten by init
                    tol=1e-7,
                    polish=False,
                    init=np.array(init)
                )
        res.approximate_error = False
        res.spectrum = optres.x.view(dtype=np.complex128)
        res.trace_error = self.trace_error(res.spectrum)
        res.approximate_error = False
        return res.spectrum


class NMRetriever(NLORetriever):
    """ This retriever uses the gradient-free Nelder-Mead algorithm.
    """
    method = "nm"

    def _retrieve(self):
        # local rename
        o = self.options
        res = self._result
        # store current guess in attribute
        spectrum = self.initial_guess.copy()
        # This algorithm is not robust against the scaling of the input vector!
        spectrum /= np.abs(spectrum).max()
        x0 = spectrum.view(np.float64).copy()
        # calculate the maximum number of function evaluations
        max_nfev = None
        if o.maxfev is not None:
            max_nfev = o.maxfev
        optres = optimize.minimize(
                    self._scalar_objective,
                    x0,
                    method='Nelder-Mead',
                    options={'maxfev': max_nfev},
                )
        res.approximate_error = False
        res.spectrum = optres.x.view(dtype=np.complex128)
        res.trace_error = self.trace_error(res.spectrum)
        res.approximate_error = False
        return res.spectrum


class BFGSRetriever(NLORetriever):
    """ This retriever uses the BFGS algorithm with numerical differentiation.
    """
    method = "bfgs"

    def _retrieve(self):
        # local rename
        o = self.options
        res = self._result
        # store current guess in attribute
        spectrum = self.initial_guess.copy()
        # This algorithm is not robust against the scaling of the input vector!
        spectrum /= np.abs(spectrum).max()
        x0 = spectrum.view(np.float64).copy()
        # calculate the maximum number of function evaluations
        max_nfev = None
        if o.maxfev is not None:
            max_nfev = o.maxfev // x0.shape[0]
        optres = optimize.minimize(
                    self._scalar_objective,
                    x0,
                    method='BFGS',
                    options={'maxiter': max_nfev},
                )
        res.approximate_error = False
        res.spectrum = optres.x.view(dtype=np.complex128)
        res.trace_error = self.trace_error(res.spectrum)
        res.approximate_error = False
        return res.spectrum



""" This module implements an object for dealing with two-dimensional data.
"""
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from . import lib
from . import io


class MeshData(io.IO):
    _io_store = ["data", "axes", "labels", "units", "uncertainty"]

    def __init__(self, data,  *axes, uncertainty=None, labels=None,
                 units=None):
        """ Creates a MeshData instance.

        Parameters
        ----------
        data : ndarray
            A at least two-dimensional array containing the data.
        *axes : ndarray
            Arrays specifying the coordinates of the data axes. Must be given
            in indexing order.
        uncertainty : ndarray
            An ndarray of the same size as `data` that contains some measure
            of the uncertainty of the meshdata. E.g., it could be the standard
            deviation of the data.
        labels : list of str, optional
            A list of strings labeling the axes. The last element labels the
            data itself, e.g. ``labels`` must have one more element than the
            number of axes.
        units : list of str, optional
            A list of unit strings.
        """
        self.data = data.copy()
        self.axes = [np.array(a).copy() for a in axes]
        if uncertainty is not None:
            self.uncertainty = uncertainty.copy()
        else:
            self.uncertainty = None
        if self.ndim != len(axes):
            raise ValueError("Number of supplied axes is wrong!")
        if self.shape != tuple(ax.size for ax in self.axes):
            raise ValueError("Shape of supplied axes is wrong!")
        self.labels = labels
        if self.labels is None:
            self.labels = ["" for ax in self.axes]
        self.units = units
        if self.units is None:
            self.units = ["" for ax in self.axes]

    @property
    def shape(self):
        """ Returns the shape of the data as a tuple.
        """
        return self.data.shape

    @property
    def ndim(self):
        """ Returns the dimension of the data as integer.
        """
        return self.data.ndim

    def copy(self):
        """ Creates a copy of the MeshData instance. """
        return MeshData(self.data, *self.axes, uncertainty=self.uncertainty,
                        labels=self.labels, units=self.units)

    def marginals(self, normalize=False, axes=None):
        """ Calculates the marginals of the data.

        axes specifies the axes of the marginals, e.g., the axes on which the
        sum is projected.
        """
        return lib.marginals(self.data, normalize=normalize, axes=axes)

    def normalize(self):
        """ Normalizes the maximum of the data to 1.
        """
        self.scale(1.0 / self.data.max())

    def scale(self, scale):
        if self.uncertainty is not None:
            self.uncertainty *= scale
        self.data *= scale

    def autolimit(self, *axes, threshold=1e-2, padding=0.25):
        """ Limits the data based on the marginals.
        """
        if len(axes) == 0:
            # default: operate on all axes
            axes = list(range(self.ndim))
        marginals = lib.marginals(self.data)
        limits = []
        for i, j in enumerate(axes):
            limit = lib.limit(self.axes[j], marginals[j],
                              threshold=threshold, padding=padding)
            limits.append(limit)
        self.limit(*limits, axes=axes)

    def limit(self, *limits, axes=None):
        """ Limits the data range of this instance.

        Parameters
        ----------
        *limits : tuples
            The data limits in the axes as tuples. Has to match the dimension
            of the data or the number of axes specified in the `axes`
            parameter.
        axes : tuple or None
            The axes in which the limit is applied. Default is `None` in which
            case all axes are selected.
        """
        if axes is None:
            # default: operate on all axes
            axes = list(range(self.ndim))
        axes = lib.as_list(axes)
        if len(axes) != len(limits):
            raise ValueError("Number of limits must match the specified axes!")
        slices = []
        for j in range(self.ndim):
            if j in axes:
                i = axes.index(j)
                ax = self.axes[j]
                x1, x2 = limits[i]
                # do it this way as we cannot assume them to be sorted...
                idx1 = np.argmin(np.abs(ax - x1))
                idx2 = np.argmin(np.abs(ax - x2))
                if idx1 > idx2:
                    idx1, idx2 = idx2, idx1
                elif idx1 == idx2:
                    raise ValueError('Selected empty slice along axis %d!' % i)
                slices.append(slice(idx1, idx2 + 1))
            else:
                # empty slice
                slices.append(slice(None))
            self.axes[j] = self.axes[j][slices[-1]]
        self.data = self.data[(*slices,)]
        if self.uncertainty is not None:
            self.uncertainty = self.uncertainty[(*slices,)]

    def interpolate(self, axis1=None, axis2=None, degree=2, sorted=False):
        """ Interpolates the data on a new two-dimensional, equidistantly
        spaced grid.
        """
        axes = [axis1, axis2]
        for i in range(self.ndim):
            if axes[i] is None:
                axes[i] = self.axes[i]
        # FITPACK can only deal with strictly increasing axes
        # so sort them beforehand if necessary...
        orig_axes = self.axes
        data = self.data.copy()
        if self.uncertainty is not None:
            uncertainty = self.uncertainty.copy()
        if not sorted:
            for i in range(len(orig_axes)):
                idx = np.argsort(orig_axes[i])
                orig_axes[i] = orig_axes[i][idx]
                data = np.take(data, idx, axis=i)
                if self.uncertainty is not None:
                    uncertainty = np.take(uncertainty, idx, axis=i)
        dataf = RegularGridInterpolator(tuple(orig_axes), data,
                                        bounds_error=False, fill_value=0.0)
        grid = lib.build_coords(*axes)
        self.data = dataf(grid)
        self.axes = axes
        if self.uncertainty is not None:
            dataf = RegularGridInterpolator(tuple(orig_axes), uncertainty,
                                            bounds_error=False, fill_value=0.0)
            self.uncertainty = dataf(grid)

    def flip(self, *axes):
        """ Flips the data on the specified axes.
        """
        if len(axes) == 0:
            return
        axes = lib.as_list(axes)
        slices = [slice(None) for ax in self.axes]
        for ax in axes:
            self.axes[ax] = self.axes[ax][::-1]
            slices[ax] = slice(None, None, -1)
        self.data = self.data[slices]
        if self.uncertainty is not None:
            self.uncertainty = self.uncertainty[slices]



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  3 16:18:05 2018

@author: Etien & Patr & Benjm

"""

"""
TEMPLATE FOR CREATION OF NEW FUNCTIONS, CLASSES, OR METHODS:
    
"""

#%% Imported modules
import numpy as np
import csv
import math
import scipy.constants as sc
import matplotlib.pyplot as plt
import scipy as sp
from scipy.optimize import curve_fit


#%% Set constants and recurrant functions
C = sc.c                          # Speed of light
pi = sc.pi                        # Pi
from numpy import sqrt       # Square root
from numpy import log         # Natural logarithm
from numpy import exp         # Exponential

#%% Methods and classes

def ezfft(t, S, normalization = "ortho", neg = False):
    """ 
    Description: Returns the Fourier transform of S and the frequency vector associated with it
    
    Inputs:
        - t: array_like
            Time vector [seconds]
        - S: array_like
            Signal vector [arb.u.]
        - normalization: str, optional
            Type of normalization for fft algorithm. For more info, see https://docs.scipy.org/doc/numpy/reference/routines.fft.html#module-numpy.fft
        - neg: bool, optional
            Choose whether or not to include negative frequencies components in output.
    
    Outputs:
        - f: ndarray
            Frequency vector [Hz]
        - y: complex ndarray
            Complex amplitude vector [arb.u.]
    """
    y = np.fft.fft(S,norm=normalization)
    y = np.fft.fftshift(y)
    f = np.fft.fftfreq(t.shape[-1], d = t[2]-t[1])
    f = np.fft.fftshift(f)
    if neg == False:
        y = 2*y[f>0]
        f = f[f>0]

    return f,y


def ezifft(f, y, normalization = "ortho"):
    """
    Description: Returns the inverse Fourier transform of y and the time vector associated with it.
    
    Inputs:
        - f: array_like
            Frequency vector [Hz]
        - y: array_like
            Complex amplitude vector [arb. u.]
        - normalization: str, optional
            Type of normalization for fft algorithm. For more info, see https://docs.scipy.org/doc/numpy/reference/routines.fft.html#module-numpy.fft
    
    Outputs:
        - t: array_like
            Time vector [seconds]
        - S: array_like
            Signal vector [arb.u.] 
    
    Other comments: Negative frequencies components should be included in f and y vectors. Add them manually if needed. 
    """
    N = len(f)
    tstep = 1/(N*(f[2]-f[1]))
    x = np.linspace(-(N*tstep/2),(N*tstep/2),N)
    y = np.fft.ifftshift(y)
    S = np.fft.ifft(y,norm=normalization)

    return x,S




def ezsmooth(x, window_len=11, window='flat'):
     """
     Description: Smooth the data using a window with requested size.
         This method is based on the convolution of a scaled window with the signal.
         The signal is prepared by introducing reflected copies of the signal
         (with the window size) in both ends so that transient parts are minimized
         in the begining and end part of the output signal.

     Inputs:
         - x: array_like
             Input signal to smooth
         - window_len: int, optional
             The dimension of the smoothing window; should be an odd integer
         - window: str, optional
             The type of window used, can be any from 'flat', 'hanning', 'hamming', 'bartlett' or 'blackman'.
             "flat" window will produce a moving average smoothing.

     Outputs:
         - y: ndarray
             Smoothed signal
     """

     if x.ndim != 1:
         raise ValueError("smooth only accepts 1 dimension arrays.")

     if x.size < window_len:
         raise ValueError("Input vector needs to be bigger than window size.")


     if window_len<3:
         return x
     
     if (window_len % 2) == 0:
        window_len+=1


     if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
         raise ValueError("Window is one of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")


     s=np.r_[x[window_len-1:0:-1],x,x[-2:-window_len-1:-1]]

     if window == 'flat': #moving average
         w=np.ones(window_len,'d')
     else:
         w=eval('np.'+window+'(window_len)')

     y=np.convolve(w/w.sum(),s,mode='valid')
     return  y[int(np.ceil(window_len/2-1)):-int(np.ceil(window_len/2-1))]


def ezcorr(x, y1, y2, unbiased=False, Norm=False, Mean = False):

    '''
    Description: Returns the correlation between two vectors y1 and y2. Returns the autocorrelation if y1 = y2.
    
    Inputs: 
        - x: array_like
            time_like vector [arb.u.]
        - y1: array_like
            First signal vector [arb.u.]
        - y2: array_like
            Second signal vector [arb.u.]
        - Bias: bool
            Method for removing the bias
        - Norm: bool
            'Normalized' or 'coef' method 
        - Mean: bool
            Option to remove the mean of the signal before correlation
            
    This function assumes the x arrays are the same, "unbiased" = True for an unbiased calculation and "Norm" = False to not normalize
    One can not have "Norm" = True and "unbiased" = False. For more information on unbiasing types, see: https://www.mathworks.com/help/matlab/ref/xcorr.html
    '''
    
    if Mean is True:
        y1 = y1-y1.mean()
        y2 = y2-y2.mean()

    delta_t = x[1]-x[0]
    ord_corr = np.correlate(y1,y2,"same")*delta_t
    absc_corr = delta_t*np.linspace(-len(ord_corr)/2,len(ord_corr)/2,len(ord_corr))
    if unbiased == True:
        if Norm == True:
            ord_unbiased = np.empty(len(absc_corr))
            for k in range(len(ord_corr)):
                ord_unbiased[k] = ord_corr[k]/(len(ord_corr)-abs(k-int(len(ord_corr)/2)))
            return(absc_corr,ord_unbiased)
        else:
            ord_unbiased = np.empty(len(absc_corr))
            for k in range(len(ord_corr)):
                ord_unbiased[k] = ord_corr[k]/(len(ord_corr)-abs(k-int(len(ord_corr)/2)))/delta_t
            return(absc_corr,ord_unbiased)
    else:
        return(absc_corr,ord_corr)


def ezcsvload(filename, nbrcolumns = 2, delimiter = '\t', decimalcomma = False, outformat = 'array', skiprows = 0, profile = None):
    """
    Description: Function for easy loading of csv-type files. Loading parameters can be set manually,
        or instruments-specific "profiles" can be called. Returns a list of array_like data.
    
    Inputs:
        - filename: str
            Name of the file to load. If not in the same folder, include path.
        - nbrcolumns: int, optional
            Number of columns contained in the file
        - delimiter: str, optional
            Character separating elements of different columns
        - decimalcomma: bool, optional
            Indicate whether or not the file use a comma to identify decimals
        - outformat: str, optional
            If "array", convert lists containing data from each columns into numpy arrays
        - skiprows: int, optional
            If file contains a header, indicate how many rows it contains to skip them.
        - profile: str, optional
            Automatically set all other options to match file format of a specific instrument.
            
    Outputs:
        - outlist: list
            Data extracted from file. Each element in the list represents a column from the file.

    """

    # Load profile, if any is specified
    if profile is not None:
        if profile is 'HR2000':
            nbrcolumns = 2
            delimiter = '\t'
            decimalcomma = False
            skiprows = 0

        if profile is 'testfile':
            nbrcolumns = 3
            delimiter = ';'
            decimalcomma = False
            skiprows = 1
        if profile is 'OSA':
            nbrcolumns = 2
            delimiter = '\t'
            decimalcomma = True
            skiprows = 0

        # Add your own profiles here

    # Preallocate output list
    outlist = [ [] for var in range(nbrcolumns) ]

    # Load file
    with open(filename, newline='') as csvfile:
        spamreader = csv.reader(csvfile, delimiter=delimiter, quotechar='|')
        for index, row in enumerate(spamreader):

            if index >= skiprows: # Skip first "skiprows" rows of file

                for outlist_index in range(nbrcolumns):

                    if decimalcomma is True: # Convert comma to marks
                        tmp = str(row[outlist_index])
                        tmp = tmp.replace(',','.')
                        try:
                            outlist[outlist_index].append(float( tmp ))
                        except:
                            print("1 data point failed")

                    else:
                        try:
                            outlist[outlist_index].append(float(row[outlist_index]))
                        except:
                            print("1 data point failed")

    if outformat is 'array': # Convert lists of values to numpy arrays
        for outlist_index in range(nbrcolumns):
            outlist[outlist_index] = np.array(outlist[outlist_index])

    return outlist


def ezfindwidth(x, y, halfwidth = False, height = 0.5, interp_points = 1e6, pos = False):
    """
    Description: Function that finds the width of an input signal or pulse. By default, the FWHM will be returned, unless height 
                 and/or halfwidth options are changed. 
                 
    Inputs: 
        - x: array_like 
            Abscissa. Width will be defined in the same units (time, frequency, wavelength, etc.) 
        - y: array_like 
            Signal/pulse to measure [arb. u.]
        - halfwidth: bool, optional
            If True, will return the halfwidth
        - height: float, optional
            Selects height for which width is calculated
        - interp points: int, optionnal
            Function will be interpolated to this number of points if the input has less points than this number
        
    Outputs:
        - width (defined in the same units (time, frequency, wavelength, etc.))
    """
    # Ensure x is stricktly ascending
    IIsort = np.argsort(x)
    x = x[IIsort]
    y = y[IIsort]
    # Max. / min. values of y array
    ymin = np.nanmin(y)
    ymax = np.nanmax(y - ymin)

    # Normalize
    ytmp = (y - ymin) / ymax

    # Interpolate data for better accuracy, if necessary
    if interp_points > len(x):
        # Interpolate data inside search domain for better accuracy
        xinterp = np.linspace(x[0] , x[-1] ,int(interp_points))
        yinterp = np.interp(xinterp,x,ytmp)
    else:
        xinterp = x
        yinterp = y

    # Cut interpolation domain at desired height
    tmp1 = (yinterp >= height )

    # Ensure there's a uniquely defined width
    tmp2 = np.linspace(0,len(tmp1)-1,len(tmp1),dtype = int)
    tmp2 = tmp2[tmp1]
    tmp3 = tmp2[1:] - tmp2[:-1]

    # Output width, if defined
    if any(tmp3 > 1) | (len(tmp2)==0):
        width = np.nan
    else:
        width = xinterp[tmp2[-1]] - xinterp[tmp2[0]]

    # Divide by two, if desired
    if halfwidth is True:
        width /= 2
        
    return width



def ezdiff(x, y, n = 1, order = 2):
    """ 
    Description: Numerical differentiation based on centered finite-difference formulas.
    
    Inputs: 
        - x: array_like; increases monotically in constant increments dx
            abscissa
        - y: array_like
            ordonate
        - n: int
            diffirentiates 'n' times
        - order: 
            "order" parameter determines the order of the finite difference formula; 
            it must be an even number greater than 0. High values of order can help 
            precision or lead to significant numerical errors, depending on the situation.
            2 is usually a safe value.
   Outputs:
       - xtrunc: array_like
           truncated abscissa
       - deriv: array_like
           nth derivative of y with respect to x.
    """

    # X increments
    dx = x[1] - x[0]

    # Number of finite difference coefficients to calculate
    nbr_coeff = 2 * math.floor( (n+1)/2 ) - 1 + order

    # p number (see wikipedia for more info)
    p = int((nbr_coeff - 1)/2)

    # Matrix of linear system Ax = b to solve
    Amatrix = np.zeros((nbr_coeff,nbr_coeff))
    for jj in range(nbr_coeff):
        tmp = -p + jj
        for kk in range(nbr_coeff):
            Amatrix[kk,jj] = tmp**kk

    # b vector
    bvector = np.zeros(nbr_coeff)
    bvector[n] = math.factorial(n)

    # Solve to find "x" vector, not related to "x" input
    # (sorry if this confusing)
    xvector = np.linalg.solve(Amatrix,bvector)

    # Preallocation
    deriv = np.zeros_like(x[p:-p])
    c = np.zeros_like(x[p:-p])

    # Evaluating finite difference, using Neumaier's improved Kahan summation
    # algorithm. Using it should reduce numerical errors during summation,
    # although it slows down calculations
    for ll in range(-p,p+1):

        new_term = (xvector[ll+p] * y[p+ll : len(y)-p+ll])/(dx**n)

        tmp1 = deriv + new_term


        cond = np.abs(deriv) >= np.abs(new_term)
        not_cond = np.logical_not(cond)

        c[cond] = c[cond] + (deriv[cond] - tmp1[cond]) + new_term[cond]

        c[(not_cond)] = c[(not_cond)] + (new_term[(not_cond)] - tmp1[(not_cond)]) + deriv[(not_cond)]


        deriv = tmp1

    # Requested derivative, corrected
    deriv += c

    # Appropriately truncated x vector for math and plotting
    xtrunc = x[p:-p]


    return xtrunc, deriv

def knife_edge_experiment(z = None, P = None, P0 = 0, P_max = None, plot = True):
    
    """
    Description: With the measured parameters from a knife edge experiment, finds the beam size
                 by fitting a error function on the data. The function on which the data is fitted is the following:
                 P0 + 0.5*P_max*( 1 - erf(sqrt(2)*(z -z0)/w) )
                 z0 and w are the fitting parameters. P_max is also a fitting parameter if not specified as input.
    
    Inputs: 
        - z: array_like
            vector containing the positions of the stage [mm]
        - P: array_like
            Singal (power) vector corresponding to the stage positions [mW]
        - P0: float, optionnal
            Signal (power) in absence of the incident beam
        - P_max: float, optional
            Maximum signal (power). i.e. when the beam is not blocked. 
            If P_max is not entered, the function will use it as another
            fitting parameter and display the computed value.
        - plot: bool, optionnal
            Plots the measured and fitted Power curves if True
    Outputs:
        - params: array
            contains in that order: z0, w, P_max (if not specified as input)
    """
    
    if z is None or P is None:
        print('Position or power array is missing')
        return
    
    if type(z) != np.ndarray or type(P) != np.ndarray:
        print('z and P must be numpy arrays')
        return
    
    if len(P) != len(z):
        print('z and P must be the same length')
        return
    
    # Defining error function
    if P_max:
        def func(z, z0, w):
            return P0 + 0.5*P_max*(1-sp.special.erf(np.sqrt(2) * (z -z0)/w))
        # Fit an error function on the data
        params, param_covar = curve_fit(func, z, P)
        z_fit = np.linspace(0.8*np.min(z), 1.2*np.max(z), 100)
        P_fit = func(z_fit, params[0], params[1])
    else:
        def func(z, z0, w, P_max):
            return P0 + 0.5*P_max*(1-sp.special.erf(np.sqrt(2) * (z -z0)/w))
        # Fit an error function on the data
        params, param_covar = curve_fit(func, z, P, bounds=(0, [np.max(z), np.max(z), np.max(P)]))
        z_fit = np.linspace(0.8*np.min(z), 1.2*np.max(z), 100)
        P_fit = func(z_fit, params[0], params[1], params[2])
        print('The fitted max power is: ' + str(params[2]))
    
    print('The beam diameter (1/e^2) is: ' + str(2*params[1]) + ' mm')
    
    # Plotting
    if plot is True:
        import femtoQ.plotting as fqp
        fqp.set_default_values_presentation()
        plt.figure()
        plt.plot(z,P, 'o', label = 'Measured')
        plt.plot( z_fit, P_fit, label = 'Fitted')
        plt.xlabel('Razor edge position (mm)')
        plt.ylabel('Average power (mW)')
        if P_max:
            plt.ylim([0,1.05*P_max])
        else:
            plt.ylim([0,1.05*params[2]])
        plt.legend()
    
        plt.show()
    
    # Returns
    return params
    
def ezpad(x, y, left, right, values=(50,50)):
    """
    Input :
    - x is the abscissa to pad, it is assumed to be a linear list
    - y is the ordinate that will be padded
    - left is the number of values to add to the left
    - right is the number of values to add to the right
    - values is (value to add to the left, value to add to the right)
    Output :
    Padded x and y lists
    """
    y = np.pad(y, (left, right), 'constant', constant_values=values) # Use numpy pad function to pad y array
    step = x[1] - x[0]
    # Manually padding the x vector assuming it is linear
    x = np.append(np.linspace(x[0]-left*step, x[0]-step, left), x)
    x = np.append(x, np.linspace(x[-1]+step, x[-1]+right*step, right))
    return x,y

class Pulse:
    """
    Description: Class to model simple electromagnetic pulse propagation and interactions.
        The pulse is modled by its electric field in the time domain.
        
    Input:
        ----------------------------------------------------------
        - lambda0: float
            Center wavelength of pulse [meters]
        - tau_FWHM: float
            FWHM pulse duration [seconds]    
        OR
        
        - v0: float
            Center frequency of pulse [Hertz]
        - v_bandwidth: float
            FWHM bandwidth in frequency domain [Hertz]
        OR
          
        - E: ndarray
            Electric field-like vector, defined in time domain [arb. u.]
        - t: ndarray
            Time vector [seconds]
        -----------------------------------------------------------
        &
        
        - A: float, optional
            Maximum amplitude of E field, when E is not inputed [arb. u.]
        - T: float, optional
            Time window represented by time vector, when t is not inputed [seconds]
        - dt: float, optional
            Resolution of the time vector, when t is not inputed [seconds]
    
    Attributes:
        - E: ndarray
            Electric field-like vector, defined in time domain. Is either inputed or calculated.
        - t: ndarray
            Time vector. Is either inputed or defined as a linear grid of resolution dt and length T
    
    Methods:
        - disperse: Returns a Pulse object on which effects of dispersion have been applied, either "manually" or through linear propagation modeling
        
        - getFWHM: Returns FWHM of either: pulse's intensity in time, power spectrum in frequency domain or power spectrum in wavelength domain
        
        - getInstFreq: Returns instantaneous angular frequency as a function of time
        
        - SFG: Returns a Pulse object calculated as a perfectly phase-matched SFG interaction between two Pulse objects.
                Allows to estimate SFG spectrum relative shape.
      
        - delay: Returns a Pulse object which has been delayed in time by a fixed amount.
    """
    def __init__(self, lambda0 = None, tau_FWHM = None, v0 = None, v_bandwidth = None, A = 1, t = None, E = None, T = 10000e-15, dt = 0.1e-15):
        #%% Simulation parameters
        self.T = T
        self.dt = dt
        if ((t is not None) & (E is not None)):
            self.t = t
            self.E = E

        elif ((lambda0 is not None) & (tau_FWHM is not None)):

            # Other pulse parameters
            tau = tau_FWHM / sqrt(2*log(2)) # Gaussian pulse half width @ e^-2
            v0 = C/lambda0                  # Pulse's carrier frequency
            w0 = 2*pi*v0                    # Pulse's carrier angular frequency

            # Electric field in time domain
            self.t = np.linspace(-T/2,T/2, round(T/self.dt) )
            self.E = A*np.exp(1j * w0 * self.t) * np.exp(-  (self.t)**2 / (tau)**2)

        elif ((v0 is not None) & (v_bandwidth is not None)):
            tauFWHM = 0.44/v_bandwidth         # Initial pulse duration, full width at half maximum
            tau = tauFWHM / sqrt(2*log(2)) # Gaussian pulse half width @ e^-2
            w0 = 2*pi*v0                    # Pulse's carrier angular frequency
            self.t = np.linspace(-T/2,T/2, round(T/self.dt) )
            self.E = A* exp(1j * w0 * (self.t)) * np.exp( - (self.t)**2 / (tau)**2)    # Electric field of the pulse in the time domain

    def disperse(self, medium = None, L = None, dispVec = None, v0 = "Auto"):
        """
        Description: This method takes the pulse and retrieves the temporal shape of the electric field
            after propagation in a given medium, using Sellmeier's equations of this medium. If an
            optionnal manual_GGD is entered, the code will not use "medium" and "L".

        Inputs:
            - material: str
                The optical medium through which the pulse propagates
            - L: float
                The thickness of material through which the pulse propagates [meters]
            - dispVec: array-like, optional
                Custom amount of dispersion. Vector containing orders of dispersion from 2 up to whatever is needed. 
                Example input is [GDD, TOD, ...], with GDD in fs^2, TOD in fs^3 and so forth.
            - v0: float or "Auto"
                Central frequency [Hertz]. If "Auto", will be calculated as average frequency of power spectrum.
        
        Outputs:
            - new_Pulse: Pulse object
                Dispersed pulse
        """
        # v is freqnecy vector, s is frequency-domain electric field vector
        v,s = ezfft(self.t,self.E, neg = True)
        # Calculate mean frequency
        if dispVec is not None:
            if v0 == "Auto":
                v0 = np.trapz(np.abs(s)**2 * v)/np.trapz(np.abs(s)**2)
            # conversion in angular frequency
            w0 = 2*pi*v0
            w = 2*pi*v
            # Calculate all phase orders
            phase = np.zeros_like(v)
            for i, disp in enumerate(dispVec):
                phase += 1/math.factorial(i+2)*disp*(w-w0)**(i+2)*(1e-15)**(i+2)
            s = s*exp(1j*phase)
        else:
            if medium.lower() == "bk7":
                lambda_1 = 0.3e-6   # Cutoff wavelengths of the equation's validity
                lambda_2 = 2.5e-6
                n_sellmeier = lambda x: (1+1.03961212/(1-0.00600069867/x**2)+0.231792344/(1-0.0200179144/x**2)+1.01046945/(1-103.560653/x**2))**.5  #https://refractiveindex.info/?shelf=glass&book=BK7&page=SCHOTT
            elif ((medium.lower() == "fs") or (medium.lower() =="fused silica")):
                lambda_1 = 0.21e-6
                lambda_2 = 6.7e-6
                n_sellmeier = lambda x: (1+0.6961663/(1-(0.0684043/x)**2)+0.4079426/(1-(0.1162414/x)**2)+0.8974794/(1-(9.896161/x)**2))**.5       #https://refractiveindex.info/?shelf=glass&book=fused_silica&page=Malitson
            elif medium.lower() == "yag":
                lambda_1 = 0.4e-6
                lambda_2 = 5e-6
                n_sellmeier = lambda x: (1+2.28200/(1-0.01185/x**2)+3.27644/(1-282.734/x**2))**.5   #https://refractiveindex.info/?shelf=main&book=Y3Al5O12&page=Zelmon
            elif medium.lower() == "sf11":
                lambda_1 = 0.37e-6
                lambda_2 = 2.5e-6
                n_sellmeier = lambda x: (1+1.73759695/(1-0.013188707/x**2)+0.313747346/(1-0.0623068142/x**2)+1.89878101/(1-155.23629/x**2))**.5      #https://refractiveindex.info/tmp/data/glass/schott/N-SF11.html
            elif medium.lower() == "sf10":
                lambda_1 = 0.38e-6
                lambda_2 = 2.5e-6
                n_sellmeier = lambda x: (1+1.62153902/(1-0.0122241457/x**2)+0.256287842/(1-0.0595736775/x**2)+1.64447552/(1-147.468793/x**2))**.5    #https://refractiveindex.info/?shelf=glass&book=SF10&page=SCHOTT
            elif medium.lower() == "znse":
                lambda_1 = 0.54e-6
                lambda_2 = 18.2e-6
                n_sellmeier = lambda x: (1+4.45813734/(1-(0.200859853/x)**2)+0.467216334/(1-(0.391371166/x)**2)+2.89566290/(1-(47.1362108/x)**2))**.5    #https://refractiveindex.info/?shelf=main&book=ZnSe&page=Connolly
            else:
                print("The entered medium does not exist or its Sellmeier's equations are not contained in this method")
                return self

            #%% Dispersive propagation
            # Convert Sellmeier's equations cutoff wavelengths to frequencies
            v1 = C / lambda_2
            v2 = C / lambda_1

            # Convert frequencies within the equation's validity domain to wavelengths in um
            vtmp = v[( (v>v1) & (v<v2) )]
            lambdatmp = C / vtmp

            # Calculate refractive index for relevant frequencies
            ntmp = n_sellmeier(lambdatmp*1e6)

            # Generate refractive index vector n(v). Values outside of validity range are
            # set to zero.
            n = np.ones_like(v)
            n[( (v>v1) & (v<v2) )] = ntmp

            # Apply material's spectral phase to frequency domain electric field
            s[v!=0] = s[v!=0] * np.exp(1j * 2 * pi * n[v!=0] * (v[v!=0] / C) * L)

        #%% Inverse fast fourier transform
        # E2 is final time domain electric field vector, t2 is its corresponding time
        # vector. If all works as intended, t2 is equivalent to t at this point
        t2, E2 = ezifft(v,s)

        #%% Final pulse realignment
        # Find final pulse's peak in time
        tpeak = np.mean( t2[np.abs(E2)**2 == np.nanmax(np.abs(E2)**2)] )

        tExtended = np.concatenate((t2-np.abs(np.min(t2)) - np.max(t2) - self.dt, t2, t2+np.abs(np.min(t2)) + np.max(t2) + self.dt), axis = 0 )

        EExtended = np.pad(E2, E2.shape, mode = 'wrap')

        # Shift time vector
        E2 = np.interp(t2, tExtended-tpeak, EExtended)

        # New object
        new_Pulse = Pulse(t = t2,E = E2)
        return new_Pulse

    def getFWHM(self, domain = "wavelength"):
        """
        Description: Returns the FWHM of the pulse, defined in either of these 3 domains, as selected by the user:
            - wavelength (default)
            - frequency
            - time
            
        Inputs:
            - domain: str, optional
                Domain in which to evaluate FWHM
                
        Outputs:
            - FWHM: float
        """
        v, s = ezfft(self.t, self.E)
        if domain == "time":
            return ezfindwidth(self.t, np.abs(self.E)**2)
        elif domain == "frequency":
            return ezfindwidth(v, np.abs(s)**2)
        elif domain == "wavelength":
            l = C/v
            return ezfindwidth(l, np.abs(s)**2)
        else:
            print("Invalid domain")
            return None


    def getInstFreq(self):
        """
        Description:  Returns instantaneous angular frequency as a function of time.
        
        No input whatsoever.
        
        Outputs:
            - w_inst: ndarray
                Instantanious angular frequency
        """
        # Temporal phase
        phi_t = np.unwrap(np.angle(self.E))
        # Instantaneous angular frequency
        w_inst = (phi_t[2:] - phi_t[:-2]) / (self.t[2] - self.t[0])
        # Add zeros the the extremities to keep the same vector length
        w_inst = np.insert(w_inst, 0, 0)
        w_inst = np.append(w_inst, 0)
        # Return instentanious frequency
        return w_inst

    def SFG(self, Pulse2, SFGonly = True):
        """
        Description: Returns the spectral amplitude of the SFG from Pulse1 and Pulse2, along with a new Pulse object
            defined by the total field resulting of the interaction. If SFGonly is False, the two SHGs will also
            be returned along with the SFG in both the field and the spectral amplitude.
            
        Inputs:
            - Pulse2: Pulse object.
                Other Pulse2 involved in SFG process. Input self for SHG.
            - SFGonly: bool
                If false, self-SHG of each input pulse will also be computed and be added to SFG field.
                
        Outputs:
            - SFG pulse: Pulse object
            - f_nl: ndarray
                Frequency vector for SFG spectrum [Hertz]
            - A_nl: ndarray
                Complex amplitude spectrum [arb. u.]
        """
        # Verification that the time vectors are the same
        if self.t.all() == Pulse2.t.all():
            # Consider SFG without SHGs
            if SFGonly is True:
                Atmp = self.E * Pulse2.E*1j
            # Consider both SHGs and the SFG
            elif SFGonly is False:
                E_tot =  self.E + Pulse2.E
                Atmp = E_tot * E_tot*1j

            f_nl, A_nl = ezfft(self.t, Atmp)
            return Pulse(t = self.t, E = Atmp), f_nl, A_nl
        else:
            print("time vector must be identical")
            return None


    def delay(self, timeDelay):
        """
        Description: Delay the pulse in time domain by a value equal to timeDelay. Does not wrap it back once it reaches the
            end of time vector t. A positive delay means the pulse's peak will shift towards negative time values.
            
        Inputs:
            - timeDelay: Time duration with which to delay the pulse [seconds]
            
        Outputs:
            - newPulse: Pulse object
                Delayed pulse.
        """
        
        newEfield = np.interp(self.t, self.t - timeDelay, self.E)
        
        newPulse = Pulse(t = self.t, E = newEfield)
        
        return newPulse













""" This sub-package implements different pulse retrieval algorithms
using a common interface. This facilitates direct comparison.

All algorithms are implemented as a subclass of :class:`BaseRetriever`. The
algorithms which are implemented step-by-step, i.e., do not rely on some
monolithic minimization algorithm implemented elsewhere, are further
subclassed from :class:`StepRetriever`.

The public function to instantiate the correct retriever is :func:`Retriever`.

This sub-package does not implement any form of data pre-processing. It expects
correctly interpolated data in form of a MeshData object.
"""
from .retriever import Retriever
from .step_retriever import (COPRARetriever, PCGPARetriever, GPARetriever,
                             GPDSCANRetriever, PIERetriever)
from .nlo_retriever import LMRetriever, NMRetriever, DERetriever, BFGSRetriever


""" This module implements the Fourier transforms on linear grids.

The following code approximates the continuous Fourier transform (FT) on
equidistantly spaced grids. While this is usually associated with
'just doing a fast Fourier transform (FFT)', surprisingly, much can be done
wrong.

The reason is that the correct expressions depend on the grid location. In
fact, the FT can be calculated with one DFT but in general it requires a prior
and posterior multiplication with phase factors.

The FT convention we are going to use here is the following::

    Ẽ(w) = 1/2pi ∫ E(t) exp(+i w t) dt
    E(t) =       ∫ Ẽ(w) exp(-i t w) dw

where w is the angular frequency. We can approximate these integrals by their
Riemann sums on the following equidistantly spaced grids::

    t_k = t_0 + k Δt, k=0, ..., N-1
    w_n = w_0 + n Δw, n=0, ..., N-1

and define E_k = E(t_k) and Ẽ_n = Ẽ(w_n) to obtain::

    Ẽ_n = Δt/2pi ∑_k E_k exp(+i w_n t_k)
    E_k = Δw     ∑_n Ẽ_n exp(-i t_k w_n).

To evaluate the sum using the FFT we can expand the exponential to obtain::

    Ẽ_n = Δt/2pi exp(+i n t_0 Δw) ∑_k [E_k exp(+i t_k w_0) ] exp(+i n k Δt Δw)
    E_k = Δw     exp(-i t_k w_0)  ∑_n [Ẽ_n exp(-i n t_0 Δw)] exp(-i k n Δt Δw)

Additionally, we have to require the so-called reciprocity relation for
the grid spacings::

          !
    Δt Δw = 2pi / N = ζ     (reciprocity relation)

This is what enables us to use the DFT/FFT! Now we look at the definition of
the FFT in NumPy::

     fft[x_m] -> X_k =     ∑_m exp(-2pi i m k / N)
    ifft[X_k] -> x_m = 1/N ∑_k exp(+2pi i k m / N)

which gives the final expressions::

    Ẽ_n = Δt N/2pi r_n   ifft[E_k s_k  ]
    E_k = Δw       s_k^*  fft[Ẽ_n r_n^*]

    with r_n = exp(+i n t_0 Δw)
         s_k = exp(+i t_k w_0)

where ^* means complex conjugation. We see that the array to be transformed
has to be multiplied with an appropriate phase factor before and after
performing the DFT. And those phase factors mainly depend on the starting
points of the grids: w_0 and t_0. Note also that due to our sign convention
for the FT we have to use ifft for the forward transform and vice versa.

Trivially, we can see that for ``w_0 = t_0 = 0`` the phase factors vanish and
the FT is approximated well by just the DFT. However, in optics these
grids are unusual.
For ``w_0 = l Δw`` and ``t_0 = m Δt``, where l, m are integers (i.e., w_0 and
t_0 are multiples of the grid spacing), the phase factors can be
incorperated into the DFT. Then the phase factors can be replaced by circular
shifts of the input and output arrays.

This is exactly what the functions (i)fftshift are doing for one specific
choice of l and m, namely for::

    t_0 = -floor(N/2) Δt
    w_0 = -floor(N/2) Δw.

In this specific case only we can approximate the FT by::

    Ẽ_n = Δt N/2pi fftshift(ifft(ifftshift(E_k)))
    E_k = Δw       fftshift( fft(ifftshift(Ẽ_n))) (no mistake!)

We see that the ifftshift _always_ has to appear on the inside. Failure to do
so will still be correct for even N (here fftshift is the same as ifftshift)
but will produce wrong results for odd N.

Additionally you have to watch out not to violate the assumptions for the
grid positions. Using a symmetrical grid, e.g.,::

    x = linspace(-1, 1, 128)

will also produce wrong results, as the elements of x are not multiples of the
grid spacing (but shifted by half a grid point).

The main drawback of this approach is that circular shifts are usually far more
time- and memory-consuming than an elementwise multiplication, especially for
higher dimensions. In fact I see no advantage in using the shift approach at
all. But for some reason it got stuck in the minds of people and you find the
notion of having to re-order the output of the DFT everywhere.

Long story short: here we are going to stick with multiplying the correct
phase factors. The code tries to follow the notation used above.

Good, more comprehensive expositions of the issues above can be found in
[Briggs1995]_ and [Hansen2014]_. For the reason why the first-order
approximation to the Riemann integral suffices, see [Trefethen2014]_.
"""
import numpy as np
# scipy.fftpack is still faster than numpy.fft (should change in numpy 1.17)
import scipy.fftpack as fft
from . import io
from .lib import twopi, sqrt2pi
_fft_backend = 'scipy'
try:
    import pyfftw
    _fft_backend = 'pyfftw'
except ImportError:
    pass


class FourierTransformBase(io.IO):
    """ This class implements the Fourier transform on linear grids.

    This simple implementation is mainly for educational use.

    Attributes
    ----------
    N : int
        Size of the grid
    dt : float
        Temporal spacing
    dw : float
        Frequency spacing (angular frequency)
    t0 : float
        The first element of the temporal grid
    w0 : float
        The first element of the frequency grid
    t : 1d-array
        The temporal grid
    w : 1d-array
        The frequency grid (angular frequency)
    """
    _io_store = ['N', 'dt', 'dw', 't0', 'w0']

    def __init__(self, N, dt=None, dw=None, t0=None, w0=None):
        """ Creates conjugate grids and calculates the Fourier transform.

        Parameters
        ----------
        N : int
            Array size
        dt : float, optional
            The temporal grid spacing. If ``None`` will be calculated by the
            reciprocity relation ``dt = 2 * pi / (N * dw)``. Exactly one of
            ``dt`` or ``dw`` has be provided.
        dw : float, optional
            The spectral grid spacing. If ``None`` will be calculated by the
            reciprocity relation ``dw = 2 * pi / (N * dt)``. Exactly one of
            ``dt`` or ``dw`` has be provided.
        t0 : float, optional
            The first element of the temporal grid. If ``None`` will be
            ``t0 = -floor(N/2) * dt``.
        w0 : float, optional
            The first element of the spectral grid. If ``None`` will be
            ``w0 = -floor(N/2) * dw``.
        """
        if dw is None and dt is not None:
            dw = np.pi / (0.5 * N * dt)
        elif dt is None and dw is not None:
            dt = np.pi / (0.5 * N * dw)
        else:
            raise ValueError("Exactly one of the grid spacings has to be "
                             "provided!")

        if t0 is None:
            t0 = -np.floor(0.5 * N) * dt
        if w0 is None:
            w0 = -np.floor(0.5 * N) * dw
        self.N = N
        self.dt = dt
        self.dw = dw
        self.t0 = t0
        self.w0 = w0
        self._post_init()

    def _post_init(self):
        """ Hook to initialize an object from storage.
        """
        # calculate the grids
        n = k = np.arange(self.N)
        self.t = self.t0 + k * self.dt
        self.w = self.w0 + n * self.dw
        # pre-calculate the phase factors
        # TODO: possibly inaccurate for large t0, w0
        self._fr = self.dt * self.N / twopi * np.exp(1.0j * n * self.t0 *
                                                     self.dw)
        self._fs = np.exp(1.0j * self.t * self.w0)
        # complex conjugate of the above
        self._br = np.exp(-1.0j * n * self.t0 * self.dw)
        self._bs = self.dw * np.exp(-1.0j * self.t * self.w0)

    def forward_at(self, x, w):
        """ Calculates the forward Fourier transform of `x` at the
        frequencies `w`.

        This function calculates the Riemann sum directly and has quadratic
        runtime. However, it can evaluate the integral at arbitrary
        frequencies, even if they are non-equidistantly spaced. Effectively,
        it performs a trigonometric interpolation.
        """
        Dnk = self.dt / twopi * np.exp(1.0j * w[:, None] * self.t[None, :])
        return Dnk @ x

    def backward_at(self, x, t):
        """ Calculates the backward Fourier transform of `x` at the
        times `t`.

        This function calculates the Riemann sum directly and has quadratic
        runtime. However, it can evaluate the integral at arbitrary
        times, even if they are non-equidistantly spaced. Effectively,
        it performs a trigonometric interpolation.
        """
        Dkn = self.dw * np.exp(-1.0j * t[:, None] * self.w[None, :])
        return Dkn @ x


# =============================================================================
# Fourier backend selection
# =============================================================================
if _fft_backend == "scipy":
    class FourierTransform(FourierTransformBase):

        def forward(self, x, out=None):
            """ Calculates the (forward) Fourier transform of ``x``.

            For n-dimensional arrays it operates on the last axis, which has
            to match the size of `x`.

            Parameters
            ----------
            x : ndarray
                The array of which the Fourier transform will be calculated.
            out : ndarray or None, optional
                A location into which the result is stored. If not provided or
                None, a freshly-allocated array is returned.
            """
            if out is None:
                out = np.empty(x.shape, dtype=np.complex128)
            out[:] = self._fr * fft.ifft(self._fs * x)
            return out

        def backward(self, x, out=None):
            """ Calculates the backward (inverse) Fourier transform of ``x``.

            For n-dimensional arrays it operates on the last axis, which has
            to match the size of `x`.

            Parameters
            ----------
            x : ndarray
                The array of which the Fourier transform will be calculated.
            out : ndarray or None, optional
                A location into which the result is stored. If not provided or
                None, a freshly-allocated array is returned.
            """
            if out is None:
                out = np.empty(x.shape, dtype=np.complex128)
            out[:] = self._bs * fft.fft(self._br * x)
            return out

elif _fft_backend == "pyfftw":
    class FourierTransform(FourierTransformBase):

        def _post_init(self):
            super()._post_init()
            # do not need the additional N factor
            n = np.arange(self.N)
            self._fr = self.dt / twopi * np.exp(1.0j * n * self.t0 * self.dw)
            # create the aligned arrays
            a = self._field = pyfftw.empty_aligned(self.N, dtype="complex128")
            b = self._spectrum = pyfftw.empty_aligned(self.N,
                                                      dtype="complex128")
            # instantiate the FFTW objects
            self._fft = pyfftw.FFTW(b, a, direction="FFTW_FORWARD")
            self._ifft = pyfftw.FFTW(a, b, direction="FFTW_BACKWARD")

        def forward(self, x, out=None):
            """ Calculates the (forward) Fourier transform of ``x``.

            For n-dimensional arrays it operates on the last axis, which has
            to match the size of `x`.

            Parameters
            ----------
            x : ndarray
                The array of which the Fourier transform will be calculated.
            out : ndarray or None, optional
                A location into which the result is stored. If not provided or
                None, a freshly-allocated array is returned.
            """
            if out is None:
                out = np.empty(x.shape, dtype=np.complex128)
            f, s = self._field, self._spectrum
            if x.ndim == 1:
                # fast code path for single dimension
                f[:] = x
                f *= self._fs
                self._ifft.execute()
                s *= self._fr
                out[:] = s
            else:
                # implicitly work along last axis and return copy
                for idx in np.ndindex(x.shape[:-1]):
                    f[:] = x[idx]
                    f *= self._fs
                    self._ifft.execute()
                    s *= self._fr
                    out[idx] = s
            return out

        def backward(self, x, out=None):
            """ Calculates the backward (inverse) Fourier transform of ``x``.

            For n-dimensional arrays it operates on the last axis, which has
            to match the size of `x`.

            Parameters
            ----------
            x : ndarray
                The array of which the Fourier transform will be calculated.
            out : ndarray or None, optional
                A location into which the result is stored. If not provided or
                None, a freshly-allocated array is returned.
            """
            if out is None:
                out = np.empty(x.shape, dtype=np.complex128)
            f, s = self._field, self._spectrum
            if x.ndim == 1:
                # fast code path for single dimension
                s[:] = x
                s *= self._br
                self._fft.execute()
                f *= self._bs
                out[:] = f
            else:
                # implicitly work along last axis and return copy
                for idx in np.ndindex(x.shape[:-1]):
                    s[:] = x[idx]
                    s *= self._br
                    self._fft.execute()
                    f *= self._bs
                    out[idx] = f
            return out


class Gaussian:
    """ This class can be used for testing the Fourier transform.
    """

    def __init__(self, dt, t0=0.0, phase=0.0):
        """ Instantiates a shifted Gaussian function.

        The Gaussian is calculated by::

            f(t) = exp(-0.5 (t - t0)^2 / dt^2) * exp(1.0j * phase)

        Its Fourier transform is::

            F(w) = dt/sqrt(2pi) exp(-0.5 * (w + phase)^2 * dt^2 +
                                     1j * t0 * w)

        Parameters
        ----------
        dt : float
            The standard deviation of the temporal amplitude distribution.
        t0 : float
            The center of the temporal amplitude distribution.
        phase : float
            The linear phase coefficient of the temporal distribution.
        """
        self.dt = dt
        self.t0 = t0
        self.phase = phase

    def temporal(self, t):
        """ Returns the temporal distribution.
        """
        arg = (t - self.t0) / self.dt
        return np.exp(-0.5 * arg**2) * np.exp(1.0j * self.phase * t)

    def spectral(self, w):
        """ Returns the spectral distribution.
        """
        w = w + self.phase
        arg = w * self.dt
        return (self.dt * np.exp(-0.5 * arg**2) * np.exp(1.0j * self.t0 * w) /
                sqrt2pi)



""" This module provides classes to calculate parametrized nonlinear process
spectra (PNPS), such as frequency-resolved optical gating (FROG),
interferometric FROG (iFROG), dispersion scan (d-scan), time-domain
ptychography (TDP) and pulse-shaper assisted methods such as multiphoton
intrapulse interference phase scan (MIIPS).

The code follows the notation used in [Geib2019]_ and its supplement.

Currently only the abovementioned methods are implemented. But the code is
written in such way that including new pulse measurement methods is very easy.
If it is a method using a collinear nonlinearity, subclass from
`CollinearPNPS`, otherwise from `NoncollinearPNPS`.

In the collinear case only `self.mask(parameter)` has to be implemented which
calculates the used linear parametrization operator. In the non-collinear
case the function `_calculate` has to be implemented which calculates and
returns the PNPS trace ``T_mn`` and the PNPS signal ``S_mk``.
"""
import numpy as np
from . import lib
from . import io
from .mesh_data import MeshData
from .frequencies import convert
from .pulse import Pulse

# global dictionary that contains all PNPS classes
_PNPS_CLASSES = {}


# =============================================================================
# Metaclass and factory
# =============================================================================
class MetaPNPS(type):
    """ Metaclass that registers PNPS classes in a global dictionary.
    """
    def __new__(cls, clsname, bases, attrs):
        global _PNPS_CLASSES
        newclass = super().__new__(cls, clsname, bases, attrs)
        processes, method = newclass._supported_processes, newclass.method
        if processes is None or method is None:
            return newclass
        # register the PNPS method name, e.g., FROG or MIIPS
        if method not in _PNPS_CLASSES:
            _PNPS_CLASSES[method] = {}
        dct = _PNPS_CLASSES[method]
        if isinstance(processes, str):
            processes = [processes]
        for p in processes:
            if p in dct:
                raise ValueError("%s-%s has two implementing classes!" %
                                 (p, method))
            dct[p] = newclass
        return newclass


class MetaIOPNPS(io.MetaIO, MetaPNPS):
    # to fix metaclass conflicts
    pass


# =============================================================================
# PNPS Base class
# =============================================================================
class BasePNPS(io.IO, metaclass=MetaIOPNPS):
    """ The PNPS base class
    """
    process = None
    method = None
    _supported_processes = None
    parameter_name = ""
    parameter_unit = ""
    # io parameters
    _io_store = ['ft', 'w0', 'w', 'process']

    def __init__(self, pulse, process, **kwargs):
        self.ft = pulse.ft
        self.w0 = pulse.w0
        self.w = pulse.w
        self.process = process
        # put the keyword arguments in the local keyspace and add them
        # to the storage list.
        for key, value in kwargs.items():
            setattr(self, key, value)
            self._add_to_storage(key)
        # add the parameter name to storage
        if self.parameter_name != "":
            self._add_to_storage(self.parameter_name)
        self._post_init()

    def _post_init(self):
        if (self._supported_processes is not None and
                self.process not in self._supported_processes):
            raise ValueError("Nonlinear process `%s` is not supported." %
                             self.process)
        # calculate the fundamental wavelength
        self.wl = convert(self.w + self.w0, "om", "wl")
        # calculate the wavelength of the process spectrum
        if self.process == "shg":
            self.process_w = 2 * self.w0 + self.ft.w
        elif self.process == "thg":
            self.process_w = 3 * self.w0 + self.ft.w
        elif self.process == "sd":
            self.process_w = self.w0 + self.ft.w
        elif self.process == "pg":
            self.process_w = self.w0 + self.ft.w
        elif self.process == "tg":
            self.process_w = self.w0 + self.ft.w
        self.process_wl = convert(self.process_w, "om", "wl")
        # store intermediate results in a dictionary
        self._tmp = dict()

    @property
    def scheme(self):
        return self.process + "-" + self.method

    def measure(self, Sk):
        """ Simulates the measurement process.

        Note that we deal with the spectrum over the frequency!
        For retrieving from actual data we need to rescale this by lambda^2.
        """
        Sn = self.ft.forward(Sk)
        return lib.abs2(Sn)

    def calculate(self, spectrum, parameter):
        """ Calculates the PNPS signal S_mk and the PNPS trace T_mn.

        Parameters
        ----------
        spectrum : 1d-array
            The pulse spectrum for which the PNPS trace is calculated.
        parameter : scalar or 1d-array
            The PNPS parameter (array) for which the PNPS trace is calculated.

        Returns
        -------
        1d- or 2d-array
            Returns the calculated PNPS trace over the frequency
            ``self.process_w``. If parameter was a scalar a 1d-array is
            returned. If it was a 1d-array a 2d-array is returned where the
            parameter runs along the first axis and the frequency along the
            second.
        """
        parameter = np.atleast_1d(parameter)
        Tmn = np.zeros((parameter.size, spectrum.size))
        Smk = np.zeros((parameter.size, spectrum.size), dtype=np.complex128)
        for m, p in enumerate(parameter):
            Tmn[m, :], Smk[m, :] = self._calculate(spectrum, p)
        # if a scalar parameter was used, squeeze out one dimension
        Tmn = Tmn.squeeze()
        Smk = Smk.squeeze()
        # store for later use (in self.trace)
        self.Tmn = Tmn
        self.Smk = Smk
        self.parameter = parameter
        self.spectrum = spectrum
        return Tmn

    def intermediate(self, parameter):
        """ Returns intermediate results as stored by the instance.
        """
        parameter = np.atleast_1d(parameter)
        tup = self._tmp[parameter[0]]
        for i, p in enumerate(parameter):
            tup = self._tmp[p]
            if i == 0:
                res = [np.zeros((parameter.shape[0], t.size), dtype=t.dtype)
                       for t in tup]
            for j, r in enumerate(res):
                r[i, :] = tup[j]
        return (r.squeeze() for r in res)

    def gradient(self, Smk2, parameter):
        """ Calculates the gradient ∇_n Z_m.
        """
        parameter = np.atleast_1d(parameter)
        Smk2 = np.atleast_2d(Smk2)
        gradnZm = np.zeros((parameter.shape[0], Smk2.shape[1]),
                           dtype=np.complex128)
        for m, p in enumerate(parameter.flat):
            gradnZm[m, :] = self._gradient(Smk2[m, :], p)
        # if a scalar parameter is passed, squeeze out one dimension
        return gradnZm.squeeze()

    @property
    def trace(self):
        """ Returns the last calculated trace as a MeshData object.
        """
        return MeshData(self.Tmn, self.parameter, self.process_w,
                        labels=[self.parameter_name, "frequency"],
                        units=[self.parameter_unit, "Hz"])


# =============================================================================
# Collinear PNPS methods
# =============================================================================
class CollinearPNPS(BasePNPS):
    """ Implements collinear methods: d-scan, iFROG, etc.
    """
    _supported_processes = ["shg", "thg", "sd"]

    def _get_tmp(self, parameter):
        if parameter in self._tmp:
            return self._tmp[parameter]
        N = self.ft.N
        Hn = self.mask(parameter)
        Ck = np.zeros(N, dtype=np.complex128)
        Sk = np.zeros(N, dtype=np.complex128)
        Tn = np.zeros(N, dtype=np.float64)
        self._tmp[parameter] = Hn, Ck, Sk, Tn
        return Hn, Ck, Sk, Tn

    def _calculate(self, spectrum, parameter):
        """ Calculates the nonlinear process spectrum for a single parameter.

        Follows the notation from our paper.
        """
        ft = self.ft
        Hn, Ck, Sk, Tn = self._get_tmp(parameter)
        ft.backward(Hn * spectrum, out=Ck)
        if self.process == "shg":
            Sk[:] = Ck * Ck
        elif self.process == "thg":
            Sk[:] = Ck * Ck * Ck
        elif self.process == "sd":
            Sk[:] = lib.abs2(Ck) * Ck
        Tn[:] = self.measure(Sk)
        return Tn, Sk

    def _gradient(self, Sk2, parameter):
        """ Returns the gradient of Z based on the previous call to _spectrum.
        """
        ft = self.ft
        # retrieve the intermediate results
        Hn, Ck, Sk, Tn = self._get_tmp(parameter)
        # difference between the updated PNPS signal and the original one
        dSk = Sk2 - Sk
        # calculate the gradients as described in the supplement
        if self.process == "shg":
            gradnZ = 2 * Hn.conj() * ft.forward(dSk * Ck.conj())
        elif self.process == "thg":
            gradnZ = 3 * Hn.conj() * ft.forward(dSk * (Ck * Ck).conj())
        elif self.process == "sd":
            gradnZ = Hn.conj() * ft.forward(dSk.conj() * Ck * Ck +
                                            2 * dSk * lib.abs2(Ck))
        # common scale for all gradients (note the minus)
        gradnZ *= -2.0 * lib.twopi * ft.dw / ft.dt
        return gradnZ


class MIIPS(CollinearPNPS):
    """ Implements the multiphoton intrapulse interference phase scan method
    (MIIPS) [Lozovoy2004]_ [Xu2006]_.

    """
    method = "miips"
    parameter_name = "delta"
    parameter_unit = "rad"

    def __init__(self, pulse, process, alpha, gamma):
        """ Creates the instance.

        Parameters
        ----------
        pulse : Pulse instance
            The pulse object that defines the simulation grid.
        process : str
            The nonlinear process used in the PNPS method.
        alpha : float
            The amplitude of the phase pattern (in rad).
        gamma : float
            The frequency of the phase pattern in Hz.
        """
        super().__init__(pulse, process, alpha=alpha, gamma=gamma)

    def mask(self, delta):
        w = self.ft.w + self.w0
        return np.exp(1.0j * self.alpha * np.cos(self.gamma * w - delta))


class IFROG(CollinearPNPS):
    """ Implements the interferometric frequency-resolved optical gating
    method [1]_.

    .. [1] G. Stibenz and G. Steinmeyer, "Interferometric frequency-resolved
            optical gating," Opt. Express 13, 2617-2626 (OSA, 2005).
    """

    method = "ifrog"
    parameter_name = "tau"
    parameter_unit = "s"

    def __init__(self, pulse, process):
        """ Creates the instance.

        Parameters
        ----------
        pulse : Pulse instance
            The pulse object that defines the simulation grid.
        process : str
            The nonlinear process used in the PNPS method.
        """
        super().__init__(pulse, process)

    def mask(self, tau):
        w = self.ft.w + self.w0
        return 0.5 + 0.5 * np.exp(-1.0j * w * tau)


class DSCAN(CollinearPNPS):
    """ Implements the dispersion scan method [Miranda2012a]_ [Miranda2012b]_.

    Not implemented in the public version of the code. Please contact us
    if you want to use pypret for d-scan measurements.
    """
    method = "dscan"
    parameter_name = "insertion"
    parameter_unit = "m"

    def __init__(self, pulse, process, material):
        raise NotImplementedError("Not implemented in public version of the "
                                  "code.")


# =============================================================================
# Noncollinear PNPS methods
# =============================================================================
class NoncollinearPNPS(BasePNPS):
    """ Implements non-collinear methods: FROG, TDP, etc.
    """
    pass


class FROG(NoncollinearPNPS):
    """ Implements frequency-resolved optical gating [Kane1993]_
    [Trebino2000]_.
    """
    _supported_processes = ["shg", "pg", "tg"]
    method = "frog"
    parameter_name = "delay"
    parameter_unit = "s"

    def __init__(self, pulse, process):
        """ Creates the instance.

        Parameters
        ----------
        pulse : Pulse instance
            The pulse object that defines the simulation grid.
        process : str
            The nonlinear process used in the PNPS method.
        """
        super().__init__(pulse, process)

    def _get_tmp(self, parameter):
        if parameter in self._tmp:
            return self._tmp[parameter]
        delay = np.exp(1.0j * parameter * self.ft.w)
        N = self.ft.N
        Ak = np.zeros(N, dtype=np.complex128)
        Ek = np.zeros(N, dtype=np.complex128)
        Sk = np.zeros(N, dtype=np.complex128)
        Tn = np.zeros(N, dtype=np.float64)
        self._tmp[parameter] = delay, Ak, Ek, Sk, Tn
        return delay, Ak, Ek, Sk, Tn

    def _calculate(self, spectrum, parameter):
        """ Calculates the nonlinear process spectrum for a single parameter.

        Follows the notation from our paper.
        """
        ft = self.ft
        delay, Ak, Ek, Sk, Tn = self._get_tmp(parameter)
        ft.backward(delay * spectrum, out=Ak)
        ft.backward(spectrum, out=Ek)
        if self.process == "shg":
            Sk[:] = Ak * Ek
        elif self.process == "pg":
            Sk[:] = lib.abs2(Ak) * Ek
        elif self.process == "tg":
            Sk[:] = Ak.conj() * (Ek * Ek)
        Tn[:] = self.measure(Sk)
        return Tn, Sk

    def _gradient(self, Sk2, parameter):
        """ Returns the gradient of Z based on the previous call to _spectrum.
        """
        ft = self.ft
        # retrieve the intermediate results
        delay, Ak, Ek, Sk, Tn = self._tmp[parameter]
        # difference between original and updated PNPS signal
        dSk = Sk2 - Sk
        # calculate the gradients as described in the supplement
        if self.process == "shg":
            gradnZ = (delay.conj() * ft.forward(dSk * Ek.conj()) +
                      ft.forward(dSk * Ak.conj()))
        elif self.process == "pg":
            gradnZ = (2 * delay.conj() *
                      ft.forward(Ak * np.real(dSk * Ek.conj())) +
                      ft.forward(dSk * lib.abs2(Ak)))
        elif self.process == "tg":
            gradnZ = (delay.conj() * ft.forward(dSk.conj() * Ek * Ek) +
                      2 * ft.forward(dSk * Ek.conj() * Ak))
        # common scale for all gradients (note the minus)
        gradnZ *= -2.0 * lib.twopi * ft.dw / ft.dt
        return gradnZ


class TDP(NoncollinearPNPS):
    """ Implements a variant of time-domain ptychography. This version is
    self-referenced and works like FROG except that in one arm of the
    correlator the bandwidth of the pulse is heavily filtered [Witting2016]_.
    Other variants are not directly supported by this class.
    """
    _supported_processes = ["shg"]
    method = "tdp"
    parameter_name = "delay"
    parameter_unit = "s"

    def __init__(self, pulse, process, center, width):
        """ Creates the instance.

        Parameters
        ----------
        pulse : Pulse instance
            The pulse object that defines the simulation grid.
        process : str
            The nonlinear process used in the PNPS method.
        center : float
            The center wavelength of the bandwidth filter in m.
        width : float
            The width (FWHM) of the bandwidth filter in m.
        """
        super().__init__(pulse, process, center=center, width=width)

    def _get_tmp(self, parameter):
        if parameter in self._tmp:
            return self._tmp[parameter]
        # convert intensity fwhm to amplitude std deviation
        sigma = 0.5 * self.width / np.sqrt(np.log(2))
        delay = (np.exp(1.0j * parameter * self.ft.w) *
                 lib.gaussian(self.wl, x0=self.center, sigma=sigma))
        N = self.ft.N
        Ak = np.zeros(N, dtype=np.complex128)
        Ek = np.zeros(N, dtype=np.complex128)
        Sk = np.zeros(N, dtype=np.complex128)
        Tn = np.zeros(N, dtype=np.float64)
        self._tmp[parameter] = delay, Ak, Ek, Sk, Tn
        return delay, Ak, Ek, Sk, Tn

    def _calculate(self, spectrum, parameter):
        """ Calculates the nonlinear process spectrum for a single parameter.

        Follows the notation from our paper.
        """
        ft = self.ft
        delay, Ak, Ek, Sk, Tn = self._get_tmp(parameter)
        ft.backward(delay * spectrum, out=Ak)
        ft.backward(spectrum, out=Ek)
        Sk[:] = Ak * Ek
        Tn[:] = self.measure(Sk)
        return Tn, Sk

    def _gradient(self, Sk2, parameter):
        """ Returns the gradient of Z based on the previous call to _spectrum.
        """
        ft = self.ft
        # retrieve the intermediate results
        delay, Ak, Ek, Sk, Tn = self._tmp[parameter]
        # difference between original and updated PNPS signal
        dSk = Sk2 - Sk
        # calculate the gradients as described in the supplement
        gradnZ = (delay.conj() * ft.forward(dSk * Ek.conj()) +
                  ft.forward(dSk * Ak.conj()))
        # common scale for all gradients (note the minus)
        gradnZ *= -2.0 * lib.twopi * ft.dw / ft.dt
        return gradnZ


# =============================================================================
# Factory method
# =============================================================================

def PNPS(pulse: Pulse, method: str, process: str, **kwargs) -> BasePNPS:
    """ Creates a PNPS instance.

    Parameters
    ----------
    pulse : Pulse
        A pulse instance that is used to simulate the PNPS trace.
    method : str
        The type of PNPS measurement. Should be one of
            - 'frog'    :class:`(see here) <pypret.pnps.FROG>`
            - 'tdp'     :class:`(see here) <pypret.pnps.TDP>`
            - 'dscan'   :class:`(see here) <pypret.pnps.DSCAN>`
            - 'miips'   :class:`(see here) <pypret.pnps.MIIPS>`
            - 'ifrog'   :class:`(see here) <pypret.pnps.IFROG>`

    process : str
        The nonlinear process used in the measurement method. Can be one of
            - 'shg' : second harmonic generation
            - 'thg' : third harmonic generation
            - 'sd' : self-diffraction
            - 'pg' : polarization gating

        Not all methods support all nonlinear processes. In that case a
        ValueError will be raised.

    Additional parameters are described in the documentation of the specific
    PNPS methods.
    """
    method = method.lower()
    process = process.lower()
    try:
        cls = _PNPS_CLASSES[method][process]
    except KeyError:
        raise ValueError("PNPS method '%s-%s' is unknown!" % (process, method))
    return cls(pulse, process=process, **kwargs)



""" This module provides the basic classes for the pulse retrieval algorithms.
"""
import numpy as np
from types import SimpleNamespace
from .. import io
from ..mesh_data import MeshData
from ..pulse_error import pulse_error
from .. import lib
from ..pnps import BasePNPS

# global dictionary that contains all PNPS classes
_RETRIEVER_CLASSES = {}


# =============================================================================
# Metaclass and factory
# =============================================================================
class MetaRetriever(type):
    """ Metaclass that registers Retriever classes in a global dictionary.
    """
    def __new__(cls, clsmethod, bases, attrs):
        global _RETRIEVER_CLASSES
        newclass = super().__new__(cls, clsmethod, bases, attrs)
        method = newclass.method
        if method is None:
            return newclass
        # register the Retriever method, e.g. 'copra'
        if method in _RETRIEVER_CLASSES:
            raise ValueError("Two retriever classes implement retriever '%s'."
                             % method)
        _RETRIEVER_CLASSES[method] = newclass
        return newclass


class MetaIORetriever(io.MetaIO, MetaRetriever):
    # to fix metaclass conflicts
    pass


# =============================================================================
# Retriever Base class
# =============================================================================
class BaseRetriever(io.IO, metaclass=MetaIORetriever):
    """ The abstract base class for pulse retrieval.

    This class implements common functionality for different retrieval
    algorithms.
    """
    method = None
    supported_schemes = None
    _io_store = ['pnps', 'options', 'logging', 'log',
                 '_retrieval_state', '_result']

    def __init__(self, pnps, logging=False, verbose=False, **kwargs):
        self.pnps = pnps
        self.ft = self.pnps.ft
        self.options = SimpleNamespace(**kwargs)
        self._result = None
        self.logging = logging
        self.verbose = verbose
        self.log = None
        rs = self._retrieval_state = SimpleNamespace()
        rs.running = False
        if (self.supported_schemes is not None and
                pnps.scheme not in self.supported_schemes):
            raise ValueError("Retriever '%s' does not support scheme '%s'. "
                             "It only supports %s." %
                             (self.method, pnps.scheme, self.supported_schemes)
                             )

    def retrieve(self, measurement, initial_guess, weights=None,
                 **kwargs):
        """ Retrieve pulse from ``measurement`` starting at ``initial_guess``.

        Parameters
        ----------
        measurement : MeshData
            A MeshData instance that contains the PNPS measurement. The first
            axis has to correspond to the PNPS parameter, the second to the
            frequency. The data has to be the measured _intensity_ over the
            frequency (not wavelength!). The second axis has to match exactly
            the frequency axis of the underlying PNPS instance. No
            interpolation is done.
        initial_guess : 1d-array
            The spectrum of the pulse that is used as initial guess in the
            iterative retrieval.
        weights : 1d-array
            Weights that are attributed to the measurement for retrieval.
            In the case of (assumed) Gaussian uncertainties with standard
            deviation sigma they should correspond to 1/sigma.
            Not all algorithms support using the weights.
        kwargs : dict
            Can override retrieval options specified in :func:`__init__`.

        Notes
        -----
        This function provides no interpolation or data processing. You have
        to write a retriever wrapper for that purpose.
        """
        self.options.__dict__.update(**kwargs)
        if not isinstance(measurement, MeshData):
            raise ValueError("measurement has to be a MeshData instance!")
        self._retrieve_begin(measurement, initial_guess, weights)
        self._retrieve()
        self._retrieve_end()

    def _retrieve_begin(self, measurement, initial_guess, weights):
        pnps = self.pnps
        if not np.all(pnps.process_w == measurement.axes[1]):
            raise ValueError("Measurement has to lie on simulation grid!")
        # Store measurement
        self.measurement = measurement
        self.parameter = measurement.axes[0]
        self.Tmn_meas = measurement.data

        self.initial_guess = initial_guess
        # set the size
        self.M, self.N = self.Tmn_meas.shape
        # Setup the weights
        if weights is None:
            self._weights = np.ones((self.M, self.N))
        else:
            self._weights = weights.copy()
        # Retrieval state
        rs = self._retrieval_state
        rs.approximate_error = False
        rs.running = True
        rs.steps_since_improvement = 0
        # Initialize result
        res = self._result = SimpleNamespace()
        res.trace_error = self.trace_error(self.initial_guess)
        res.approximate_error = False
        res.spectrum = self.initial_guess.copy()
        # Setup the logger
        if self.logging:
            log = self.log = SimpleNamespace()
            log.trace_error = []
            log.initial_guess = self.initial_guess.copy()
        else:
            self.log = None
        if self.verbose:
            print("Started retriever '%s'" % self.method)
            print("Options:")
            print(self.options)
            print("Initial trace error R = {:.10e}".format(res.trace_error))
            print("Starting retrieval...")
            print()

    def _retrieve_end(self):
        rs = self._retrieval_state
        rs.running = False
        res = self._result
        if res.approximate_error:
            res.trace_error = self.trace_error(res.spectrum)
            res.approximate_error = False

    def _project(self, measured, Smk):
        """ Performs the projection on the measured intensity.
        """
        # in frequency domain
        Smn = self.ft.forward(Smk)
        # project and specially handle values with zero amplitude
        absSmn = np.abs(Smn)
        f = (absSmn > 0.0)
        Smn[~f] = np.sqrt(measured[~f] + 0.0j)
        Smn[f] = Smn[f] / absSmn[f] * np.sqrt(measured[f] + 0.0j)
        # back in time domain
        Smk2 = self.ft.backward(Smn)
        return Smk2

    def _objective_function(self, spectrum):
        """ Calculates the minimization objective from the pulse spectrum.

        This is Eq. 11 in the paper:

            r = sum (Tmn^meas - mu * Tmn)
        """
        # calculate the PNPS trace
        Tmn = self.pnps.calculate(spectrum, self.parameter)
        return self._r(Tmn)

    def trace_error(self, spectrum, store=True):
        """ Calculates the trace error from the pulse spectrum.
        """
        Tmn = self.pnps.calculate(spectrum, self.parameter)
        return self._R(Tmn, store=store)

    def _r(self, Tmn, store=True):
        """ Calculates the minimization objective r from a simulated trace Tmn.
        """
        diff = self._error_vector(Tmn, store=store)
        return np.sum(diff * diff)

    def _error_vector(self, Tmn, store=True):
        """ Calculates the residual vector from measured to simulated
        intensity.
        """
        # rename
        rs = self._retrieval_state
        Tmn_meas = self.Tmn_meas
        # scaling factor
        w2 = self._weights * self._weights
        mu = np.sum(Tmn_meas * Tmn * w2) / np.sum(Tmn * Tmn * w2)
        # store intermediate results in current retrieval state
        if store:
            rs.mu = mu
            rs.Tmn = Tmn
            rs.Smk = self.pnps.Smk
        return np.ravel((Tmn_meas - mu * Tmn) * self._weights)

    def _R(self, Tmn, store=True):
        """ Calculates the trace error from a simulated trace Tmn.
        """
        r = self._r(Tmn, store=store)
        return self._Rr(r)

    def _Rr(self, r):
        """ Calculates the trace error from the minimization objective r.
        """
        return np.sqrt(r / (self.M * self.N *
                            (self.Tmn_meas * self._weights).max()**2))

    def result(self, pulse_original=None, full=True):
        """ Analyzes the retrieval results in one retrieval instance
            and processes it for plotting or storage.
        """
        rs = self._retrieval_state
        if self._result is None or self._retrieval_state.running:
            return None
        res = SimpleNamespace()
        # the meta data
        res.parameter = self.parameter
        res.options = self.options
        res.logging = self.logging
        res.measurement = self.measurement
        # store the retriever itself
        if full:
            res.pnps = self.pnps
        else:
            res.pnps = None

        # the pulse spectra
        # 1 - the retrieved pulse
        res.pulse_retrieved = self._result.spectrum
        # 2 - the original test pulse, optional
        res.pulse_original = pulse_original
        # 3 - the initial guess
        res.pulse_initial = self.initial_guess

        # the measurement traces
        # 1 - the original data used for retrieval
        res.trace_input = self.Tmn_meas
        # 2 - the trace error and the trace calculated from the retrieved pulse
        res.trace_error = self.trace_error(res.pulse_retrieved)
        res.trace_retrieved = rs.mu * rs.Tmn
        res.response_function = rs.mu
        # the weights
        res.weights = self._weights

        # this is set if the original spectrum is provided
        if res.pulse_original is not None:
            # the trace error of the test pulse (non-zero for noisy input)
            res.trace_error_optimal = self.trace_error(res.pulse_original)
            # 3 - the optimal trace calculated from the test pulse
            res.trace_original = rs.mu * rs.Tmn
            dot_ambiguity = False
            if self.pnps.method == "ifrog" or self.pnps.scheme == "shg-frog":
                dot_ambiguity = True
            # the pulse error to the test pulse
            res.pulse_error, res.pulse_retrieved = pulse_error(
                    res.pulse_retrieved, res.pulse_original, self.ft,
                    dot_ambiguity=dot_ambiguity)

        if res.logging:
            # the logged trace errors
            res.trace_errors = np.array(self.log.trace_error)
            # the running minimum of the trace errors (for plotting)
            res.rm_trace_errors = np.minimum.accumulate(res.trace_errors,
                                                        axis=-1)
        if self.verbose:
            lib.retrieval_report(res)
        return res


def Retriever(pnps: BasePNPS, method: str = "copra", maxiter=300, maxfev=None,
              logging=False, verbose=False, **kwargs) -> BaseRetriever:
    """ Creates a retriever instance.

    Parameters
    ----------
    pnps : PNPS
        A PNPS instance that is used to simulate a PNPS measurement.
    method : str, optional
        Type of solver.  Should be one of
            - 'copra'       :class:`(see here) <COPRARetriever>`
            - 'gpa'         :class:`(see here) <GPARetriever>`
            - 'gp-dscan'     :class:`(see here) <GPDSCANRetriever>`
            - 'pcgpa'       :class:`(see here) <PCGPARetriever>`
            - 'pie'         :class:`(see here) <PIERetriever>`
            - 'lm'          :class:`(see here) <LMRetriever>`
            - 'bfgs'        :class:`(see here) <BFGSRetriever>`
            - 'de'          :class:`(see here) <DERetriever>`
            - 'nelder-mead' :class:`(see here) <NMRetriever>`

        'copra' is the default choice.
    maxiter : int, optional
        The maximum number of algorithm iterations. The default is 300.
    maxfev : int, optional
        The maximum number of function evaluations. If given, the algorithms
        stop before this number is reached. Not all algorithms support this
        feature. Default is ``None``, in which case it is ignored.
    logging : bool, optional
        Stores trace errors and pulses over the iterations if supported
        by the retriever class. Default is `False`.
    verbose : bool, optional
        Prints out trace errors during the iteration if supported by the
        retriever class. Default is `False`.
    """
    method = method.lower()
    try:
        cls = _RETRIEVER_CLASSES[method]
    except KeyError:
        raise ValueError("Retriever '%s' is unknown!" % (method))
    return cls(pnps, maxiter=maxiter, maxfev=maxfev,
               logging=logging, verbose=verbose, **kwargs)



""" This module implements specific pulse retrieval algorithms, e.g.,
COPRA, GPA, PCGPA, etc.
"""
import numpy as np
from scipy.optimize import minimize_scalar
from .retriever import BaseRetriever
from .. import lib


class StepRetriever(BaseRetriever):

    def _retrieve(self):
        # local rename
        o = self.options
        log = self.log
        res = self._result
        rs = self._retrieval_state
        # store current guess in attribute
        spectrum = self.initial_guess.copy()
        # initialize R
        R = res.trace_error
        for i in range(o.maxiter):
            # store trace error and spectrum for later analysis
            if self.logging:
                # if the trace error was only approximated, calculate it here
                if rs.approximate_error:
                    R = self.trace_error(spectrum, store=False)
                log.trace_error.append(R)
            # Perform a single retriever step in one of the algorithms
            R, new_spectrum = self._retrieve_step(i, spectrum.copy())
            # update the solution if the result is better
            if R < res.trace_error:
                # R is calculated for the input, i.e., the old spectrum.
                res.trace_error = R
                res.approximate_error = rs.approximate_error
                res.spectrum[:] = spectrum  # store the old spectrum
                rs.steps_since_improvement = 0
            else:
                rs.steps_since_improvement += 1
            # accept the new spectrum
            spectrum[:] = new_spectrum
            if self.verbose:
                if i == 0:
                    print("iteration".ljust(10) + "trace error".ljust(20))
                s = "{:d}".format(i + 1).ljust(10)
                if rs.approximate_error:
                    s += "~"
                else:
                    s += " "
                s += "{:.10e}".format(R)
                if R == res.trace_error:
                    s += "*"
                print(s)
            if not rs.running:
                break
        if self.verbose:
            print()
            print("~ approximate trace error")
            print("* accepted as best trace error")
            print()

        # return the retrieved spectrum
        # for a more detailed analysis call self.result()
        return res.spectrum


class COPRARetriever(StepRetriever):
    """ This module implements the common pulse retrieval algorithm
    [Geib2019]_.
    """
    method = "copra"

    def __init__(self, pnps, alpha=0.25, **kwargs):
        """ For a full documentation of the arguments see :class:`Retriever`.

        Parameters
        ----------
        alpha : float, optional
            Scales the step size in the global stage of COPRA. Higher values
            mean potentially faster convergence but less accuracy. Lower
            values provide higher accuracy for the cost of speed. Default is
            0.25.
        """
        super().__init__(pnps, alpha=alpha, **kwargs)

    def _retrieve_begin(self, measurement, initial_guess, weights):
        super()._retrieve_begin(measurement, initial_guess, weights)
        pnps = self.pnps
        rs = self._retrieval_state
        rs.mode = "local"  # COPRA starts with local mode
        # calculate the maximum gradient norm
        # self.trace_error() was called beforehand -> rs.Tmn and rs.Smk exist!
        Smk2 = self._project(self.Tmn_meas / rs.mu, rs.Smk)
        nablaZnm = pnps.gradient(Smk2, self.parameter)
        rs.current_max_gradient = np.max(np.sum(lib.abs2(nablaZnm), axis=1))

    def _retrieve_step(self, iteration, En):
        """ Perform a single COPRA step.

        Parameters
        ----------
        iteration : int
            The current iteration number - mainly for logging.
        En : 1d-array
            The current pulse spectrum.
        """
        # local rename
        ft = self.ft
        options = self.options
        pnps = self.pnps
        rs = self._retrieval_state
        Tmn_meas = self.Tmn_meas
        # current gradient -> last gradient
        rs.previous_max_gradient = rs.current_max_gradient
        rs.current_max_gradient = 0.0
        # switch iteration
        if rs.steps_since_improvement == 5:
            rs.mode = "global"
        # local iteration
        if rs.mode == "local":
            # running estimate for the trace
            Tmn = np.zeros((self.M, self.N))
            for m in np.random.permutation(np.arange(self.M)):
                p = self.parameter[m]
                Tmn[m, :] = pnps.calculate(En, p)
                Smk2 = self._project(Tmn_meas[m, :] / rs.mu, pnps.Smk)
                nablaZnm = pnps.gradient(Smk2, p)
                # calculate the step size
                Zm = lib.norm2(Smk2 - pnps.Smk)
                gradient_norm = lib.norm2(nablaZnm)
                if gradient_norm > rs.current_max_gradient:
                    rs.current_max_gradient = gradient_norm
                gamma = Zm / max(rs.current_max_gradient,
                                 rs.previous_max_gradient)
                # update the spectrum
                En -= gamma * nablaZnm
            # Tmn is only an approximation as En changed in the iteration!
            rs.approximate_error = True
            R = self._R(Tmn)  # updates rs.mu!!!
        # global iteration
        elif rs.mode == "global":
            Tmn = pnps.calculate(En, self.parameter)
            r = self._r(Tmn)
            R = self._Rr(r)  # updates rs.mu!!!
            rs.approximate_error = False
            # gradient descent w.r.t. Smk
            w2 = self._weights * self._weights
            gradrmk = (-4 * ft.dt / (ft.dw * lib.twopi) *
                       ft.backward(rs.mu * ft.forward(pnps.Smk) *
                                   (Tmn_meas - rs.mu * Tmn) * w2))
            etar = options.alpha * r / lib.norm2(gradrmk)
            Smk2 = pnps.Smk - etar * gradrmk
            # gradient descent w.r.t. En
            nablaZn = pnps.gradient(Smk2, self.parameter).sum(axis=0)
            # calculate the step size
            Z = lib.norm2(Smk2 - pnps.Smk)
            etaz = options.alpha * Z / lib.norm2(nablaZn)
            # update the spectrum
            En -= etaz * nablaZn
        return R, En


class PCGPARetriever(StepRetriever):
    """ This class implements the principal components generalized projections
    algorithm (PCGPA) for SHG-FROG.

    We follow the algorithm as described in [Kane1999]_ but use the PNPS
    formalism from [Geib2019]_ and some minor modifications:

    - it supports both the singular value decomposition and the power
      method to find/approximate the largest eigenvector.
    - the projection includes the scaling factor µ. This makes the method
      robust against initial guesses with the wrong magnitude. It should
      have no adverse effect.

    """
    method = "pcgpa"
    supported_schemes = ["shg-frog"]

    def __init__(self, pnps, decomposition="power", **kwargs):
        """ For a full documentation of the arguments see :class:`Retriever`.

        Parameters
        ----------
        decomposition : str, optional
            It specifies how the FROG signal is decomposed. If `power` (the
            default) the power method is used to find the largest eigenvalue.
            If `svd` a full singular value decomposition is performed. This
            is potentially much slower but more accurate.
        """
        super().__init__(pnps, decomposition=decomposition, **kwargs)

    def _retrieve_begin(self, measurement, initial_guess, weights):
        super()._retrieve_begin(measurement, initial_guess, weights)
        if np.any(self.parameter != measurement.axes[0]):
            raise ValueError("The delay has to be sampled exactly at the "
                             "temporal simulation grid.")

    def _retrieve_step(self, iteration, En):
        # local rename
        ft = self.ft
        options = self.options
        pnps = self.pnps
        Tmn_meas = self.Tmn_meas
        rs = self._retrieval_state

        R = self.trace_error(En)  # updates rs.mu!!!
        # project on measured intensity
        Smk2 = self._project(Tmn_meas / rs.mu, pnps.Smk)
        # to outer product form
        for n in range(ft.N):
            Smk2[:, n] = np.roll(Smk2[::-1, n], n)
        if options.decomposition == "power":
            # apply power method iteration once
            Ek = ft.backward(En)
            Ek[:] = Smk2.conj() @ Ek
            Ek[:] = Ek.conj() / lib.norm(Ek)
        elif options.decomposition == "svd":
            # use full svd (slow!)
            U, s, V = np.linalg.svd(Smk2)
            Ek = U[:, 0] * np.sqrt(s[0])  # select U
        ft.forward(Ek, out=En)
        return R, En


class GPARetriever(StepRetriever):
    """ Implements the classical generalized projections algorithm for
    SHG-FROG as described in [DeLong1994]_ and [Trebino2000].

    As far as I know the determination of the step size in GPA is not
    made explicit in the publications. It is usually done in a line search.
    In this implementation we offer three different options:

    - an exact line search using a Brent style minimizer
    - a backtracking (inexact) line search using the Armijo-Goldstein
      condition with c=0.5 and tau=0.5.
    - the same heuristic choice for the step size used in copra.

    The last method is the fastest, but as the first is the classic choice
    for GPA, it is the default.
    """
    method = "gpa"
    supported_schemes = ["shg-frog"]

    def __init__(self, pnps, step_size="exact", **kwargs):
        """ For a full documentation of the arguments see :class:`Retriever`.

        Parameters
        ----------
        step_size : str, optional
            Specifies how the step size of the gradient step in GPA is
            determined. Default is `exact` which performs an exact line search.
            `inexact` performs a backtracking line search and `copra` uses the
            ad-hoc estimates for the step size used in COPRA.
        """
        super().__init__(pnps, step_size=step_size, **kwargs)

    def _retrieve_begin(self, measurement, initial_guess):
        super()._retrieve_begin(measurement, initial_guess)
        if np.any(self.parameter != measurement.axes[0]):
            raise ValueError("The delay has to be sampled exactly at the "
                             "temporal simulation grid.")

    def _retrieve_step(self, iteration, En):
        # local rename
        ft = self.ft
        options = self.options
        pnps = self.pnps
        Tmn_meas = self.Tmn_meas
        rs = self._retrieval_state

        R = self.trace_error(En)  # updates rs.mu!!!
        # obtain intermediate results
        delay, Amk, Ek, Smk, Tmn = pnps.intermediate(self.parameter)
        Ek = Ek[0, :]  # the same for every parameter
        # project on measured intensity
        Smk2 = self._project(Tmn_meas / rs.mu, Smk)
        # calculate the gradient of Z w.r.t. to the temporal pulse envelope
        # by directly implementing (S58) of [Geib2019]_
        dS = Smk2 - Smk
        indices = np.array(np.rint(self.parameter / ft.dt), dtype=np.int32)
        gradient = np.zeros((self.M, self.N), dtype=np.complex128)
        for m in range(self.M):
            gradient[m, :] = np.roll(dS[m, :] * Ek.conj(), -indices[m])
        gradient = -2 * np.sum(gradient + dS * Amk.conj(), axis=0)

        # approximate the step size with the "copra" step size
        gamma0 = lib.norm2(dS) / lib.norm2(gradient)
        if options.step_size == "copra":
            # directly choose the copra step size
            gamma = gamma0
        else:
            # do a line search
            def objective(gamma):
                self.trace_error(ft.forward(Ek - gamma * gradient))
                return lib.norm2(Smk2 - pnps.Smk)
            if options.step_size == "exact":
                # perform an exact line search
                bracket = [0.9 * gamma0, gamma0]
                ret = minimize_scalar(objective,
                                      bracket=bracket,
                                      method="brent")
                gamma = ret.x
            elif options.step_size == "inexact":
                # perform a back-tracking line search until the Armijo
                # condition is fulfilled.
                t = 0.5 * lib.norm2(gradient)
                tau = 0.5
                if iteration == 0:
                    rs.old_gamma = 5.0 * gamma0
                gamma = 2 * rs.old_gamma
                objective0 = objective(0.0)
                while objective(gamma) - objective0 > -gamma * t:
                    gamma = gamma * tau
                rs.old_gamma = gamma
        Ek = Ek - gamma * gradient

        ft.forward(Ek, out=En)
        return R, En


class GPDSCANRetriever(StepRetriever):
    """ This class implements a pulse retrieval algorithm for SHG and THG
    d-scan based on the paper [Miranda2017]_.

    In our tests we found that it does not converge in the noiseless case.
    In other words the global solution to the least-squares problem is not a
    fixed point of the iteration.
    """
    supported_schemes = ["shg-dscan", "thg-dscan"]
    method = "gp-dscan"

    def _retrieve_begin(self, measurement, initial_guess):
        super()._retrieve_begin(measurement, initial_guess)
        pnps = self.pnps
        rs = self._retrieval_state
        # calculate phase mask once
        rs.Hmn = np.zeros((self.M, self.N), dtype=np.complex128)
        for i, p in enumerate(self.parameter):
            rs.Hmn[i, :] = pnps.mask(p)

    def _retrieve_step(self, iteration, En):
        # local rename
        ft = self.ft
        pnps = self.pnps
        Tmn_meas = self.Tmn_meas
        rs = self._retrieval_state
        Hmn = rs.Hmn

        R = self.trace_error(En)  # updates rs.mu!!!
        # project on measured intensity
        Smk2 = self._project(Tmn_meas / rs.mu, pnps.Smk)
        # modify En
        if pnps.process == "shg":
            Smk2 *= ft.backward(Hmn * En).conj()
            Smk2 /= np.abs(Smk2)**(2/3)
            En = np.sum(Hmn.conj() * ft.forward(Smk2), axis=0)
        elif pnps.process == "thg":
            Smk2 *= ft.backward(Hmn * En).conj()**2
            Smk2 /= np.abs(Smk2)**(4/5)
            En = np.sum(Hmn.conj() * ft.forward(Smk2), axis=0)
        return R, En


class PIERetriever(StepRetriever):
    """ This class implements a pulse retrieval algorithm for SHG-FROG based on
    the ptychographical iterative engine (PIE). It is based on the paper
    [Sidorenko2016]_ and its erratum [Sidorenko2017]_.

    We modified the algorithm to include the scaling factor µ in the
    projection. This makes the method robust against initial guesses with the
    wrong magnitude. It should have no adverse effect.
    """
    method = "pie"
    supported_schemes = ["shg-frog"]

    def _retrieve_step(self, iteration, En):
        # local rename
        ft = self.ft
        pnps = self.pnps
        rs = self._retrieval_state
        Tmn_meas = self.Tmn_meas

        # running estimate for the trace
        Tmn = np.zeros((self.M, self.N))
        # random choice for the step size scaling
        beta = np.random.uniform(0.1, 0.5)
        for m in np.random.permutation(np.arange(self.M)):
            p = self.parameter[m]
            Tmn[m, :] = pnps.calculate(En, p)
            # get intermediate results from private attribute
            delay, Amk, Ek, Smk, _ = pnps._tmp[p]
            # project
            Smk2 = self._project(Tmn_meas[m, :] / rs.mu, Smk)
            # perform update
            Ek += beta * Amk.conj() * (Smk2 - Smk) / lib.abs2(Ek).max()
            # update the spectrum
            ft.forward(Ek, out=En)

        # Tmn is only an approximation as En changed in the iteration!
        rs.approximate_error = True
        R = self._R(Tmn)  # updates rs.mu!!!

        return R, En



""" Provides a class to simulate an ultrashort optical pulse using its envelope
description.

The temporal envelope is denoted as `field` and the spectral envelope as
`spectrum` in the code and the function signatures.
"""
import numpy as np
from . import io
from . import lib
from .frequencies import convert
from scipy.optimize import minimize_scalar, root_scalar


class Pulse(io.IO):
    """ A class for modelling femtosecond pulses by their envelope.
    """
    _io_store = ['ft', 'wl0', '_field', '_spectrum']

    def __init__(self, ft, wl0, unit='wl'):
        """ Initializes an optical pulse described by its envelope.

        Parameters
        ----------
        ft : FourierTransform
            A ``FourierTransform`` instance that specifies a temporal and
            spectral grid.
        wl0 : float
            The center frequency of the pulse.
        unit : str
            The unit in which the center frequency is specified. Can be either
            of ``wl``, ``om``, ``f``, or ``k``. See ``frequencies`` for more
            information. Default is ``wl``.
        """
        self.ft = ft
        self.wl0 = convert(wl0, unit, 'wl')
        self._field = np.zeros(ft.N, dtype=np.complex128)
        self._spectrum = np.zeros(ft.N, dtype=np.complex128)
        self._post_init()

    def copy(self):
        """ Returns a copy of the pulse object.

        Note that they still reference the same `FourierTransform` instance,
        which is assumed to be immutable.
        """
        p = Pulse(self.ft, self.wl0)
        p.spectrum = self.spectrum
        return p

    def _post_init(self):
        ft = self.ft
        self.t = ft.t
        self.w = ft.w
        self.dt = ft.dt
        self.dw = ft.dw
        self.N = ft.N

        self.w0 = convert(self.wl0, 'wl', 'om')
        self.wl = convert(self.w + self.w0, 'om', 'wl')

    @property
    def field(self):
        """ The complex-valued temporal envelope of the pulse.

        On read access returns a copy of the internal array. On write
        access the spectral envelope is automatically updated.
        """
        return self._field.copy()

    @field.setter
    def field(self, val):
        self._field[:] = val
        self.update_spectrum()

    def field_at(self, t):
        """ The complex-valued temporal envelope of the pulse at the
        times `t`.
        """
        return self.ft.backward_at(self._spectrum, t)

    @property
    def spectrum(self):
        """ The complex-valued spectral envelope of the pulse.

        On read access returns a copy of the internal array. On write
        access the temporal envelope is automatically updated.
        """
        return self._spectrum.copy()

    @spectrum.setter
    def spectrum(self, val):
        self._spectrum[:] = val
        self.update_field()

    def spectrum_at(self, w):
        """ The complex-valued spectral envelope of the pulse at the
        frequencies `w`.
        """
        return self.ft.forward_at(self._field, w)

    def update_field(self):
        """ Manually updates the field from the (modified) spectrum.
        """
        self.ft.backward(self._spectrum, out=self._field)

    def update_spectrum(self):
        """ Manually updates the spectrum from the (modified) field.
        """
        self.ft.forward(self._field, out=self._spectrum)

    @property
    def intensity(self):
        """ The temporal intensity profile of the pulse in vacuum.

        Only read access.
        """
        return lib.abs2(self._field)

    @property
    def amplitude(self):
        """ The temporal amplitude profile of the pulse in vacuum.

        Only read access.
        """
        return self._field.abs()

    @property
    def phase(self):
        """ The temporal phase of the pulse.

        Only read access.
        """
        return lib.phase(self._field)

    @property
    def spectral_intensity(self):
        """ The spectral intensity profile of the pulse in vacuum.

        Only read access.
        """
        return lib.abs2(self._spectrum)

    @property
    def spectral_amplitude(self):
        """ The spectral amplitude profile of the pulse in vacuum.

        Only read access.
        """
        return self._spectrum.abs()

    @property
    def spectral_phase(self):
        """ The spectral phase of the pulse.

        Only read access.
        """
        return lib.phase(self._spectrum)

    @property
    def time_bandwidth_product(self):
        """ Calculates the rms time-bandwidth product of the pulse.

            In this definition a transform-limited Gaussian pulse has a
            time-bandwidth product of 0.5. So the number returned by this
            function will always be >= 0.5.
        """
        return (lib.standard_deviation(self.t, self.intensity) *
                lib.standard_deviation(self.w, self.spectral_intensity))

    def fwhm(self, dt=None):
        """ Calculates the full width at half maximum (FWHM) of the temporal
        intensity profile.

        Parameters
        ----------
        dt : float or None, optional
            Specifies the required accuracy of the calculation. If `None` (the
            default) it is only as good as the spacing of the underlying
            simulation grid - which can be quite coarse compared to the FWHM.
            If smaller it is calculated based on trigonometric interpolation.

        """
        t, intensity = self.t, self.intensity
        if dt is None or dt == self.dt:
            return lib.fwhm(t, intensity)

        # exact calculation
        def objective(tau):
            return lib.abs2(self.field_at(np.array([tau]))[0])

        # determine the maximum accurately
        idx = np.argmax(intensity)
        res = minimize_scalar(lambda x: -objective(x),
                              (t[idx] - self.dt, t[idx]),
                              tol=dt / 100.0, method="brent")
        y0 = -res.fun
        # determine the right and left sided intersection points
        idx1, idx2 = lib.arglimit(intensity, threshold=0.5 * y0,
                                  padding=0.0, normalize=False)
        # left side
        res = root_scalar(lambda x: objective(x) - 0.5 * y0,
                          bracket=(t[idx1] - self.dt, t[idx1]),
                          xtol=dt, method="brentq")
        xl = res.root
        # right side
        res = root_scalar(lambda x: objective(x) - 0.5 * y0,
                          bracket=(t[idx2], t[idx2] + self.dt),
                          xtol=dt, method="brentq")
        xr = res.root
        return xr - xl
